# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import logging
import os
import subprocess
import time


from novaclient import exceptions
from novaclient import client


from ucivms import (
    errors,
    timeouts,
    vms,
)


def uci_image_name(domain, series, architecture):
    """Returns an image name.

    The images are uploaded to glance for specific needs.

    :param domain: 'cloudimg' or 'britney'.

    :param series: The ubuntu series (precise, trusty, etc).

    :param architecture: The processor architecture ('amd64', i386, etc).
    """
    if domain not in ('cloudimg', 'britney'):
        raise ValueError('Invalid image domain')
    return 'uci/{}/{}-{}.img'.format(domain, series, architecture)


def get_os_nova_client(conf, debug=False):
    os_nova_client = client.Client(
        '1.1',
        conf.get('vm.os.username'), conf.get('vm.os.password'),
        conf.get('vm.os.tenant_name'),
        conf.get('vm.os.auth_url'),
        region_name=conf.get('vm.os.region_name'),
        service_type='compute')
    return os_nova_client


class NovaServerException(errors.UciVmsError):
    pass


class NovaClient(object):
    """A nova client re-trying requests on known transient failures."""

    def __init__(self, conf, **kwargs):
        self.logger = kwargs.pop('logger')
        self.first_wait = kwargs.pop('first_wait', 30)
        self.wait_up_to = kwargs.pop('wait_up_to', 600)
        self.retries = kwargs.pop('retries', 8)
        debug = kwargs.pop('debug', False)
        # Activating debug will output the http requests issued by nova and the
        # corresponding responses (/!\ including credentials).
        if debug:
            self.logger.setLevel(logging.DEBUG)
        self.nova = get_os_nova_client(conf, debug)

    def retry(self, func, *args, **kwargs):
        no_404_retry = kwargs.pop('no_404_retry', False)
        sleeps = timeouts.ExponentialBackoff(
            self.first_wait, self.wait_up_to, self.retries)
        for attempt, sleep in enumerate(sleeps, start=1):
            try:
                if attempt > 1:
                    self.logger.info('Re-trying {} {}/{}'.format(
                        func.__name__, attempt, self.retries))
                return func(*args, **kwargs)
            except client.requests.ConnectionError:
                # Most common transient failure: the API server is unreachable
                msg = 'Connection error for {}, will sleep for {} seconds'
                self.logger.warn(msg.format(func.__name__, sleep))
            except (exceptions.OverLimit, exceptions.RateLimit):
                msg = ('Rate limit reached for {},'
                       ' will sleep for {} seconds')
                # This happens rarely but breaks badly if not caught. elmo
                # recommended a 30 seconds nap in that case.
                sleep += 30
                self.logger.exception(msg.format(func.__name__, sleep))
            except exceptions.ClientException as e:
                if no_404_retry and e.http_status == 404:
                    raise
                msg = '{} failed will sleep for {} seconds'
                self.logger.exception(msg.format(func.__name__, sleep))
            except:
                # All other exceptions are raised
                self.logger.exception('{} failed'.format(func.__name__))
                raise NovaServerException('{} failed'.format(func.__name__))
            # Take a nap before retrying
            self.logger.info('Sleeping {} seconds for {} {}/{}'.format(
                sleep, func.__name__, attempt, self.retries))
            time.sleep(sleep)
        # Raise if we didn't succeed at all
        raise NovaServerException("Failed to '{}' after {} retries".format(
            func.__name__, attempt))

    def flavors_list(self):
        return self.retry(self.nova.flavors.list)

    def images_list(self):
        return self.retry(self.nova.images.list)

    def create_server(self, name, flavor, image, user_data, nics,
                      availability_zone):
        return self.retry(self.nova.servers.create, name=name,
                          flavor=flavor, image=image, userdata=user_data,
                          nics=nics, availability_zone=availability_zone)

    def delete_server(self, server_id):
        # FIXME: 404 shouldn't be retried, if it's not there anymore, there is
        # nothing to delete. -- vila 2015-07-16
        return self.retry(self.nova.servers.delete, server_id)

    def start_server(self, instance):
        return self.retry(instance.start)

    def stop_server(self, instance):
        return self.retry(instance.stop)

    def create_floating_ip(self):
        return self.retry(self.nova.floating_ips.create)

    def delete_floating_ip(self, floating_ip):
        return self.retry(self.nova.floating_ips.delete, floating_ip)

    def add_floating_ip(self, instance, floating_ip):
        return self.retry(instance.add_floating_ip, floating_ip)

    def get_server_details(self, server_id):
        return self.retry(self.nova.servers.get, server_id,
                          no_404_retry=True)

    def get_server_console(self, server, length=None):
        return self.retry(server.get_console_output, length)


class NovaServer(vms.VM):

    nova_client_class = NovaClient

    def __init__(self, conf, logger=None):
        super(NovaServer, self).__init__(conf)
        if logger is None:
            # FIXME: We probably want to generalize logging -- vila 2015-08-25
            self.ensure_dir(self.config_dir_path())
            logging.basicConfig(
                level=logging.INFO,
                format="%(asctime)s %(levelname)s %(message)s",
                filename=os.path.join(self.config_dir_path(), 'uci-vms.log'))
            logger = logging.getLogger()
        self.logger = logger
        self.instance = None
        self.floating_ip = None
        self.nova = self.build_nova_client()
        self.test_bed_key_path = None
        self.conf.set('vm.ssh_authorized_keys',
                      self.conf.get('vm.ssh_key_path') + '.pub')
        # No need to reboot a nova instance
        self.conf.set('vm.poweroff', 'False')
        self.conf.set('vm.final_message', 'testbed setup completed.')

    # MISSINGTEST
    def state(self):
        try:
            with open(self.nova_id_path()) as f:
                nova_id = f.read().strip()
        except IOError as e:
            # python2 does not provide FileNotFoundError
            if e.errno == errno.ENOENT:
                # Unknown interface
                return 'UNKNOWN'
        try:
            self.instance = self.nova.get_server_details(nova_id)
        except exceptions.NotFound:
            return 'UNKNOWN'
        # The instance may remain in the DELETED state for some time.
        nova_states = dict(BUILD='STARTING',
                           ACTIVE='RUNNING',
                           SHUTOFF='STOPPED',
                           DELETED='UNKNOWN')
        return nova_states[self.instance.status]

    def build_nova_client(self):
        nova_client = self.nova_client_class(self.conf, logger=self.logger)
        return nova_client

    def ensure_ssh_key_is_available(self):
        self.test_bed_key_path = self.conf.get('vm.ssh_key_path')
        # FIXME: Needs to be unified for all vm classes -- vila 2015-08-26
        # From the test runner, we need an ssh key that can be used to connect
        # to the testbed. For testing purposes, we rely on self.auth_conf to
        # provide this key.
        if not os.path.exists(self.test_bed_key_path):
            base_dir = os.path.dirname(self.test_bed_key_path)
            try:
                # Try to create needed dirs
                os.makedirs(base_dir)
            except OSError as e:
                # They are already there, no worries
                if e.errno != errno.EEXIST:
                    raise
            # First time the test runner instance needs to create a test bed
            # instance, we need to create the ssh key pair.
            subprocess.call(
                ['ssh-keygen', '-t', 'rsa', '-q',
                 '-f', self.test_bed_key_path, '-N', ''])

    def find_flavor(self):
        flavors = self.conf.get('vm.os.flavors')
        existing_flavors = self.nova.flavors_list()
        for flavor in flavors:
            for existing in existing_flavors:
                if flavor == existing.name:
                    return existing
        raise NovaServerException(
            'None of [{}] can be found'.format(','.join(flavors)))

    def find_nova_image(self):
        image_name = self.conf.get('vm.image')
        existing_images = self.nova.images_list()
        for existing in existing_images:
            if image_name == existing.name:
                return existing
        raise NovaServerException(
            'Image "{}" cannot be found'.format(image_name))

    def find_nics(self):
        net_id = self.conf.get('vm.net_id')
        if net_id:
            return [{'net-id': self.conf.get('vm.net_id')}]
        return None

    # FIXME: This should save the console whether or not the setup fails
    # -- vila 2015-08-26
    def setup(self):
        flavor = self.find_flavor()
        image = self.find_nova_image()
        nics = self.find_nics()
        self.ensure_ssh_key_is_available()
        self.create_user_data()
        with open(self._user_data_path) as f:
            user_data = f.read()
        self.instance = self.nova.create_server(
            name=self.conf.get('vm.name'), flavor=flavor, image=image,
            user_data=user_data, nics=nics,
            # FIXME: We probably want at least a vm.az_name option. And get
            # that option from higher levels too -- vila 2014-10-13
            availability_zone=None)
        self.create_nova_id_file(self.instance.id)
        self.wait_for_active_instance()
# FIXME: We want a vm.create_floating_ip option ? -- vila 2015-08-24
#        if unit_config.is_hpcloud(self.conf.get('os.auth_url')):
#            self.floating_ip = self.nova.create_floating_ip()
#            self.nova.add_floating_ip(self.instance, self.floating_ip)
        self.wait_for_ip()
        self.wait_for_cloud_init()
        self.ensure_ssh_works()
        # FIXME: inherited from uci-engine, revisit -- vila 2016-01-18
        ppas = self.conf.get('vm.ppas')
        if ppas:
            cmd = ['sudo', 'add-apt-repository']
            if self.conf.get('vm.release') > 'precise':
                cmd.append('--enable-source')
            for ppa in ppas:
                self.ssh(*(cmd + [ppa]))
        # Now we can apt-get update (doing it earlier would lead to the wrong
        # source package to be installed).
        self.safe_apt_get_update()

    def apt_get_update(self):
        return self.ssh('sudo', 'apt-get', 'update')

    def safe_apt_get_update(self):
        for timeout in self.conf.get('vm.apt_get.update.timeouts'):
            proc, out, err = self.apt_get_update()
            if proc.returncode == 0:
                # We're done
                return
            else:
                msg = ('apt-get update failed, wait {}s\n'
                       'stdout:\n{}\n'
                       'stderr:\n{}\n')
                self.logger.info(msg.format(timeout, out, err))
                time.sleep(float(timeout))
        raise NovaServerException('apt-get update never succeeded')

    def update_instance(self, nova_id=None):
        if nova_id is None:
            nova_id = self.instance.id
        try:
            # Always query nova to get updated data about the instance
            self.instance = self.nova.get_server_details(nova_id)
            return True
        except:
            # But catch exceptions if something goes wrong. Higher levels will
            # deal with the instance not replying.
            return False

    def wait_for_active_instance(self):
        timeout_limit = time.time() + self.conf.get('vm.nova.boot_timeout')
        while (time.time() < timeout_limit and
               self.instance.status not in ('ACTIVE', 'ERROR')):
            time.sleep(5)
            self.update_instance()
        if self.instance.status != 'ACTIVE':
            msg = 'Instance {} never came up (last status: {})'.format(
                self.instance.id, self.instance.status)
            raise NovaServerException(msg)

    def nova_id_path(self):
        return os.path.join(self.config_dir_path(), 'nova_id')

    def create_nova_id_file(self, nova_id):
        nova_id_path = self.nova_id_path()
        self.ensure_dir(self.config_dir_path())
        with open(nova_id_path, 'w') as f:
            f.write(nova_id + '\n')

    # FIXME: ~Duplicated with lxc, refactor to extract discover_ip() which is
    # the part where the ip is acquired with different means depending on the
    # vm provider. -- vila 2015-10-25
    def wait_for_ip(self):
        timeout_limit = time.time() + self.conf.get('vm.nova.set_ip_timeout')
        while time.time() < timeout_limit:
            if not self.update_instance():
                time.sleep(5)
                continue
            networks = self.instance.networks.values()
            if networks:
                # The network name is arbitrary, can vary for different clouds
                # but there should be only one network so we get the first one
                # and avoid the need for a config option for the network name.
                # We take the last IP address so it's either the only one or
                # the floating one. In both cases that gives us a reachable IP.
                self.ip = networks[0][-1]
                self.logger.info('Got IP {} for {}'.format(
                    self.ip, self.instance.id))
                # FIXME: Why not get it from the console ? -- vila 2015-08-26
                # MISSINGTEST
                self.create_iface_file('eth0', self.ip, 'unknown', 'unknown')
                return
            else:
                self.logger.info(
                    'IP not yet available for {}'.format(self.instance.id))
            time.sleep(5)
        msg = 'Instance {} never provided an IP'.format(self.instance.id)
        raise NovaServerException(msg)

    def get_cloud_init_console(self, length=None):
        return self.nova.get_server_console(self.instance, length)

    # FIXME: ~Duplicated with lxc, refactor to extract discovering cloud-init
    # completion (success/failure may be refined later if at least ssh access
    # is provided. In this case, the cloud-init log files can be
    # acquired/analyzed more precisely. This will require refactoring
    # setup(). -- vila 2015-10-25
    def wait_for_cloud_init(self):
        timeout_limit = (time.time() +
                         self.conf.get('vm.nova.cloud_init_timeout'))
        final_message = self.conf.get('vm.final_message')
        while time.time() < timeout_limit:
            # A relatively cheap way to catch cloud-init completion is to watch
            # the console for the specific message we specified in user-data).
            # FIXME: or at least check that we don't miss when we sleep a
            # significant time between two calls (like on canonistack where we
            # can sleep for minute(s) -- vila 2015-07-17
            console = self.get_cloud_init_console(10)
            if final_message in console:
                # We're good to go
                self.logger.info(
                    'cloud-init completed for {}'.format(self.instance.id))
                return
            time.sleep(5)
        raise NovaServerException(
            'Instance {} never completed cloud-init'.format(self.instance.id))

    def ensure_ssh_works(self):
        proc, out, err = self.ssh('whoami')
        if proc.returncode:
            msg = ('testbed {} IP {} cannot be reached with ssh, retcode: {}\n'
                   'stdout:\n{}\n'
                   'stderr:\n{}\n')
            self.logger.info(msg.format(self.instance.id, self.ip,
                                        proc.returncode, out, err))
            raise NovaServerException('No ssh access to {}, IP: {}'.format(
                self.instance.id, self.ip))

    # MISSINGTEST
    def start(self):
        self.nova.start_server(self.instance)

    # MISSINGTEST
    def stop(self):
        self.nova.stop_server(self.instance)

    def teardown(self):
        if self.instance is not None:
            self.logger.info('Deleting instance {}'.format(self.instance.id))
            self.nova.delete_server(self.instance.id)
            self.instance = None
            os.remove(self.nova_id_path())
        if self.floating_ip is not None:
            self.nova.delete_floating_ip(self.floating_ip)
            self.floating_ip = None
        # FIXME: Now we can remove the testbed key from known_hosts (see
        # ssh()). -- vila 2014-01-30

    # FIXME: Should be unified with 'shell' and 'shell_captured'
    # -- vila 2015-08-26
    def ssh(self, command, *args, **kwargs):
        """Run a command in the testbed via ssh.

        :param args: The command and its positional arguments as a list.

        FIXME: Unused ?!? -- vila 2015-10-25
        :param kwargs: The named arguments for the command.

        The stdout and stderr outputs are captured and returned to the caller.
        """
        user = 'ubuntu'
        host = self.ip
        cmd = ['ssh',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               '{}@{}'.format(user, host)]
        if command is not None:
            cmd += [command]
        if args:
            cmd += args
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        return proc, out, err

    # FIXME: Should be provided by the base class -- vila 2015-08-26
    def scp(self, local_path, remote_path):
        cmd = ['scp',
               # FIXME: It would be better to ssh-keygen -f
               # "~/.ssh/known_hosts" -R self.ip once we're done with the test
               # bed to avoid polluting ssh commands stdout, but that will do
               # for now (that's what juju is doing after all ;)
               # -- vila 2014-01-29
               '-oStrictHostKeyChecking=no',
               '-i', self.test_bed_key_path,
               local_path, remote_path]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = proc.communicate()
        return proc, out, err

    def get_remote_content(self, path):
        proc, content, err = self.ssh('cat', path)
        if proc.returncode:
            # We didn't get a proper content, report it instead
            content = ("{} couldn't be copied from testbed {}:\n"
                       "error: {}\n".format(path, self.ip, err))
        return content



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import logging
import os
import time

from ucivms import (
    errors,
    subprocesses,
    timeouts,
    vms,
)


# FIXME: Too much duplication from nova here... -- vila 2015-12-22
def uci_image_name(domain, series, architecture):
    """Returns an image name.

    Images can be exported to lxd for specific needs.

    :param domain: 'cloudimg' for example.

    :param series: The ubuntu series (trusty, xenial, etc).

    :param architecture: The processor architecture ('amd64', 'i386', etc).
    """
    return 'uci/{}/{}-{}'.format(domain, series, architecture)


def lxd_info(vm_name, source=None):
    info = dict(state='UNKNOWN')
    if source is None:
        info_cmd = ['lxc', 'info', vm_name]
        try:
            _, source, err = subprocesses.run(info_cmd)
        except errors.CommandError as e:
            if e.err == 'error: not found\n':
                return info
            else:
                raise
    scanning_ips = False
    for l in source.splitlines():
        if scanning_ips:
            if l == '(none)':
                # interfaces are not setup yet
                continue
            details = l.split('\t')
            try:
                ifce, proto, ip = details[0:3]
            except ValueError:
                raise ValueError('Unexpected IP line: {}'.format(l))
                # FIXME: Assume eth0 until we can do better -- vila 2015-12-22
            if 'eth0' in ifce and proto == 'IPV4':
                info['ip'] = ip
        elif l.startswith('Status:'):
            _, s = l.split()
            info['state'] = s.upper()
        elif l.startswith('Init:'):
            _, pid = l.split()
            info['pid'] = pid
        elif l.startswith('Ips:'):
            scanning_ips = True
    return info


class Lxd(vms.VM):

    def __init__(self, conf, logger=None):
        super(Lxd, self).__init__(conf)
        if logger is None:
            # FIXME: Duplicated from nova.py -- vila 2015-12-21
            self.ensure_dir(self.config_dir_path())
            logging.basicConfig(
                level=logging.INFO,
                format="%(asctime)s %(levelname)s %(message)s",
                filename=os.path.join(self.config_dir_path(), 'uci-vms.log'))
            logger = logging.getLogger()
        self.logger = logger

    def state(self):
        # man lxc(7) defines the possible states as: STOPPED, STARTING,
        # RUNNING, ABORTING, STOPPING. We add UNKNOWN.
        info = lxd_info(self.conf.get('vm.name'))
        return info['state']

    def init(self):
        init_cmd = ['lxc', 'init', self.conf.get('vm.image'),
                    self.conf.get('vm.name')]
        return subprocesses.run(init_cmd)

    def set_cloud_init_config(self):
        self.create_user_data()
        config_cmd = ['lxc', 'config', 'set', self.conf.get('vm.name'),
                      'user.user-data', '-']
        return subprocesses.run(config_cmd, cmd_input=self.ci_user_data.dump())

    def setup(self):
        _, out, err = self.init()
        # FIXME: Log out & err ? -- vila 2015-12-22
        _, out, err = self.set_cloud_init_config()
        network = self.conf.get('vm.network')
        if network:
            dev_add_cmd = (['lxc', 'config', 'device', 'add',
                            self.conf.get('vm.name')] + network)
            # FIXME: Log out & err ? -- vila 2016-01-05
            _, out, err = subprocesses.run(dev_add_cmd)
        return self.start()

    # MISSINGTEST
    # FIXME: duplicated from Lxc -- vila 2015-12-22
    def wait_for_ip(self):
        ip = None
        first, up_to, retries = self.conf.get('vm.lxd.set_ip_timeouts')
        first = float(first)
        up_to = float(up_to)
        retries = int(retries)
        exc = None
        for sleep in timeouts.ExponentialBackoff(first, up_to, retries):
            try:
                info = lxd_info(self.conf.get('vm.name'))
                if info['state'] != 'RUNNING' or 'ip' not in info:
                    raise errors.UciVmsError(
                        'Lxd {} has not provided an IP yet: {}'.format(
                            self.conf.get('vm.name'), info))
                ip = info['ip']
                # FIXME: Assume eth0 until we can do better -- vila 2015-12-22
                self.create_iface_file('eth0', ip, 'unknown', 'unknown')
                # Success, clear the exception
                exc = None
                break
            except (errors.CommandError, errors.UciVmsError) as e:
                exc = e
                # FIXME: logging + metric  -- vila 2015-06-25
                pass
            time.sleep(sleep)
        if exc is not None:
            # Re-raise the last seen exception
            raise exc
        if ip is None:
            raise errors.UciVmsError('Lxd {} never provided an IP'.format(
                self.conf.get('vm.name')))

    # MISSINGTEST
    # FIXME: duplicated from Lxc -- vila 2015-12-22
    def wait_for_ssh(self):
        first, up_to, retries = self.conf.get('vm.lxd.ssh_setup_timeouts')
        first = float(first)
        up_to = float(up_to)
        retries = int(retries)
        exc = None
        for attempt, sleep in enumerate(timeouts.ExponentialBackoff(
                first, up_to, retries), start=1):
            try:
                ret, out, err = self.shell_captured('whoami')
                # Success, clear the exception
                exc = None
                break
            # FIXME: There may be better feedback to provide here if (for
            # example, the IP is lost abruptly or the vm is killed for reasons)
            # -- vila 2015-10-24
            except errors.CommandError as e:
                exc = e
                # FIXME: logging + metric  -- vila 2015-06-25
                pass
            time.sleep(sleep)
        if exc is not None:
            # Re-raise the last seen exception
            raise exc

    def start(self):
        start_cmd = ['lxc', 'start', self.conf.get('vm.name')]
        _, out, err = subprocesses.run(start_cmd)
        self.wait_for_ip()
        self.wait_for_ssh()

    def stop(self):
        stop_cmd = ['lxc', 'stop', self.conf.get('vm.name')]
        return subprocesses.run(stop_cmd)

    def teardown(self):
        return subprocesses.run(
            ['lxc', 'delete', self.conf.get('vm.name')])

# MISSINGTEST: Experiment and run tests in qemu where lxd is not installed

# MISSINGTEST: for stop(), teardown()
# MISSINGTEST: for download() which NIY

# Strip down the cloud-init dependency to the bare minimum so it's easier to
# ensure it didn't break. Turn the stripped features into ssh/lxc exec
# commands.

# /etc/cloud/cloud.cfg is the defaut config for cloud-init

# /etc/cloud/cloud.cfg.d/90_dpkg.cfg may be worth overriding to avoid searching
# for clouds we don't use



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import os
import unittest

from ucitests import features
from ucivms import (
    config,
    errors,
)
from ucivms.tests import (
    features as vms_features,
    fixtures as vms_fixtures,
)
from ucivms.vms import lxd


class TestUciImageName(unittest.TestCase):

    def test_valid_cloud_image(self):
        self.assertEqual(
            'uci/cloudimg/trusty-amd64',
            lxd.uci_image_name('cloudimg', 'trusty', 'amd64'))


@features.requires(vms_features.lxd_client_feature)
class TestInit(unittest.TestCase):

    def setUp(self):
        super(TestInit, self).setUp()
        # FIXME: The following will be needed by all lxd tests
        # -- vila 2015-12-21
        # lxc requires a cert and key for the client and automagically generate
        # them on first need.
        lxd_conf_dir = os.path.expanduser('~/.config/lxc')
        vms_fixtures.isolate_from_disk(self)
        os.environ['LXD_CONF'] = lxd_conf_dir
        # To isolate tests from each other, created vms needs a unique name. To
        # keep those names legal and still user-readable we use the class name
        # and the test method name. The process id is added so that the same
        # test suite can be run on the same host in several processes.
        self.vm_name = '{}-{}-{}'.format(self.__class__.__name__,
                                         self._testMethodName, os.getpid())
        # lxc wants a valid hostname, '_' are clearly forbidden
        self.vm_name = self.vm_name.replace('_', '-')
        config_dir = os.path.join(self.uniq_dir, 'config')
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
vm.config_dir={config_dir}
[{vm_name}]
vm.name = {vm_name}
vm.class = lxd
'''.format(config_dir=config_dir, vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()
        self.addCleanup(conf.store.save_changes)

    def get_image_id(self, series, arch):
        return lxd.uci_image_name('cloudimg', series, arch)

    def test_unknown_image(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        image_id = self.get_image_id('Idontexist', 'amd64')
        vm.conf.set('vm.image', image_id)
        with self.assertRaises(errors.CommandError) as cm:
            vm.init()
        err_msg = "error: can't get info for image '{}': not found\n"
        self.assertEqual(err_msg.format(image_id),
                         cm.exception.err)

    def test_known_image(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        image_id = self.get_image_id('xenial', 'amd64')
        vm.conf.set('vm.image', image_id)
        self.addCleanup(vm.teardown)
        ret, out, err = vm.init()
        self.assertEqual(0, ret)
        self.assertEqual('Creating {} done.\n'.format(vm.conf.get('vm.name')),
                         out)
        self.assertEqual('', err)


class TestInfo(unittest.TestCase):

    def test_unknown(self):
        info = lxd.lxd_info('idontexist', '')
        self.assertEqual('UNKNOWN', info['state'])

    def test_running(self):
        info = lxd.lxd_info('foo', '''
Name: foo
Status: Running
Init: 8114
Processcount: 20
Ips:
  eth0:	IPV4	10.0.3.130	vethKJJF5J
  lo:	IPV4	127.0.0.1
  lo:	IPV6	::1
  lxcbr0:	IPV4	10.0.4.1
''')
        self.assertEqual('RUNNING', info['state'])
        self.assertEqual('8114', info['pid'])
        self.assertEqual('10.0.3.130', info['ip'])

    def test_stopped(self):
        info = lxd.lxd_info('foo', '''
Name: foo
Status: Stopped
''')
        self.assertEqual({'state': 'STOPPED'}, info)

    def test_no_ips(self):
        # Before ips are set
        info = lxd.lxd_info('foo', '''
Name: foo
Status: Running
Ips:
(none)
''')
        self.assertEqual({'state': 'RUNNING'}, info)
        self.assertFalse('ip' in info)


@features.requires(vms_features.lxd_client_feature)
class TestSetup(unittest.TestCase):

    def setUp(self):
        super(TestSetup, self).setUp()
        # FIXME: The following will be needed by all lxd tests
        # -- vila 2015-12-21
        # lxc requires a cert and key for the client and automagically generate
        # them on first need.
        lxd_conf_dir = os.path.expanduser('~/.config/lxc')
        vms_fixtures.isolate_from_disk(self)
        os.environ['LXD_CONF'] = lxd_conf_dir
        # To isolate tests from each other, created vms needs a unique name. To
        # keep those names legal and still user-readable we use the class name
        # and the test method name. The process id is added so that the same
        # test suite can be run on the same host in several processes.
        self.vm_name = '{}-{}-{}'.format(self.__class__.__name__,
                                         self._testMethodName, os.getpid())
        # lxc wants a valid hostname, '_' are clearly forbidden
        self.vm_name = self.vm_name.replace('_', '-')
        config_dir = os.path.join(self.uniq_dir, 'config')
        # Create a shared config
        conf = config.VmStack(None)
        conf.store._load_from_string('''
vm.ssh_opts=-oUserKnownHostsFile=/dev/null,-oStrictHostKeyChecking=no
# FIXME: A bit too specific for a test ;-D -- vila 2015-12-22
vm.ssh_authorized_keys = /home/vila/.ssh/keys/vila@launchpad.pub
lxd.network=eth0,nic,nictype=bridged,parent=br0,name=eth0
vm.config_dir={config_dir}
[{vm_name}]
vm.name = {vm_name}
vm.class = lxd
'''.format(config_dir=config_dir, vm_name=self.vm_name))
        conf.store.save()
        conf.store.unload()
        self.addCleanup(conf.store.save_changes)

    def get_image_id(self, series, arch):
        return lxd.uci_image_name('cloudimg', series, arch)

    def test_usable(self):
        vm = lxd.Lxd(config.VmStack(self.vm_name))
        image_id = self.get_image_id('xenial', 'amd64')
        vm.conf.set('vm.image', image_id)
        self.addCleanup(vm.teardown)
        vm.setup()
        # We should be able to ssh with the right user
        ret, out, err = vm.shell_captured('whoami')
        self.assertEqual(0, ret)
        self.assertEqual('ubuntu\n', out)



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import errno
import os

from uciconfig import (
    options,
    registries,
    stacks,
    stores,
)


class StartingNameMatcher(stacks.SectionMatcher):
    """A sub name section matcher.

    This selects sections starting with a given name respecting the Store
    order.
    """

    def __init__(self, store, name):
        super(StartingNameMatcher, self).__init__(store)
        self.name = name

    def get_sections(self):
        """Get all sections starting with ``name`` in the store.

        The most generic sections are described first in the store, then more
        specific ones can be provided for reduced scopes.

        The returned sections are therefore returned in the reversed order so
        the most specific ones can be found first.
        """
        store = self.store
        # Later sections are more specific, they should be returned first
        for _, section in reversed(list(store.get_sections())):
            if section.id is None:
                # The no-name section is always included if present
                yield store, section
                continue
            if self.name is None:
                # If no name has been defined, no section can match
                continue
            section_name = section.id
            if self.name.startswith(section_name):
                yield store, section


VmMatcher = StartingNameMatcher
VmStore = stores.FileStore
VmCmdLineStore = stores.CommandLineStore


def system_config_dir():
    return '/etc'


def config_file_basename():
    return 'uci-vms.conf'


class VmStack(stacks.Stack):
    """Per-directory options."""

    def __init__(self, name):
        """Make a new stack for a given vm.

        :param name: The name of a virtual machine.

        The options are searched in following files each one providing a
        ``name`` specific section and defaults in the no-name section.

        The directory local file:
        * the ``name`` section in ./uci-vms.conf,
        * the no-name section in ./uci-vms.conf
        The user file:
        * the ``name`` section in ~/uci-vms.conf,
        * the no-name section in ~/uci-vms.conf
        The system-wide file:
        * the ``name`` section in /etc/uci-vms.conf,
        * the no-name section in /etc/uci-vms.conf
        """
        self.cmdline_store = VmCmdLineStore()
        lpath = os.path.abspath(config_file_basename())
        self.local_store = self.get_shared_store(VmStore(lpath))
        section_getters = [self.cmdline_store.get_sections,
                           VmMatcher(self.local_store, name).get_sections]
        upath = os.path.join(os.environ['HOME'], config_file_basename())
        if upath != lpath:
            # Only add user store if we're not running from HOME
            user_store = self.get_shared_store(VmStore(upath))
            section_getters.append(VmMatcher(user_store, name).get_sections)
        else:
            user_store = self.local_store
        spath = os.path.join(system_config_dir(), config_file_basename())
        self.system_store = self.get_shared_store(VmStore(spath))
        section_getters.append(VmMatcher(self.system_store, name).get_sections)
        super(VmStack, self).__init__(
            section_getters, user_store, mutable_section_id=name)

    # FIXME: This should be a DictOption -- vila 2016-01-05
    def get_nova_creds(self):
        """Get nova credentials from a config.

        This defines the set of options needed to authenticate against nova in
        a single place.

        :raises: uciconfig.errors.OptionMandatoryValueError if one of the
            options is not set.
        """
        creds = {}
        for k in ('username', 'password', 'tenant_name',
                  'auth_url', 'region_name'):
            opt_name = 'vm.os.{}'.format(k)
            creds[opt_name] = self.get(opt_name)
        return creds


def path_from_unicode(path_string):
    return os.path.expanduser(path_string)


class PathOption(options.Option):

    def __init__(self, *args, **kwargs):
        """A path option definition.

        This possibly expands the user home directory.
        """
        super(PathOption, self).__init__(
            *args, from_unicode=path_from_unicode, **kwargs)


class PackageListOption(options.ListOption):
    """A package list option definition.

    This possibly expands '@filename' replacing it with the file content.
    """

    def __init__(self, *args, **kwargs):
        # Foce invalid values to be an error (and forbids overriding it) to
        # catch invalid file names
        super(PackageListOption, self).__init__(
            *args, invalid='error', **kwargs)

    def from_unicode(self, string):
        values = super(PackageListOption, self).from_unicode(string)
        if not values:
            return values
        converted = []
        for v in values:
            if v.startswith('@'):
                try:
                    # FIXME: A bit more care should be taken about interpreting
                    # relative paths: are they relative to the config file ?
                    # Also, the current implementation swallow the ValueError
                    # raised below, turning it into an OptionValueError
                    # instead, losing the precision if multiple files are
                    # involved -- vila 2015-06-18
                    with open(v[1:]) as f:
                        for package in f:
                            converted.append(package.strip())
                except IOError as e:
                    # python2 does not provide FileNotFoundError
                    if e.errno == errno.ENOENT:
                        raise ValueError('{} does not exist'.format(v[1:]))
                    else:
                        raise
            else:
                converted.append(v)
        return converted


def register(option):
    options.option_registry.register(option)

register(options.Option('vm', default=None,
                        help_string='''\
The name space defining a virtual machine.

This option is a place holder to document the options that defines a virtual
machine and the options defining the infrastructure used to manage them all.

For qemu based vms, the definition of a vm is stored in an xml file under
'/etc/libvirt/qemu/{vm.name}.xml'. This is under the libvirt package control
and is out of scope for uci-vms.

There are 3 other significant files used for a given vm:

- a disk image mounted at '/' from '/dev/sda1':
  '{vm.images_dir}/{vm.name}.qcow2'

- a iso image available from '/dev/sdb' labeled 'cidata':
  {vm.images_dir}/{vm.name}.seed which contains the cloud-init data used to
  configure/install/update the vm.

- a console: {vm.images_dir}/{vm.name}.console which can be 'tail -f'ed from
  the host.

The data used to create the seed above are stored in a vm specific
configuration directory for easier debug and reference:
- {vm.config_dir}/user-data
- {vm.config_dir}/meta-data
- {vm.config_dir}/ecdsa
- {vm.config_dir}/ecdsa.pub
'''))

# The directory where we store vm files related to their configuration with
# cloud-init (user-data, meta-data, ssh keys).
register(PathOption('vm.vms_dir', default='~/.config/uci-vms',
                    help_string='''\
Where vm related config files are stored.

This includes user-data and meta-data for cloud-init and ssh server keys.

This directory must exist.

Each vm get a specific directory (automatically created) there based on its
name.
'''))
# The base directories where vms are stored for kvm
register(PathOption('vm.images_dir', default='/var/lib/libvirt/images',
                    help_string='''Where vm disk images are stored.''',))
register(options.Option('vm.qemu_etc_dir', default='/etc/libvirt/qemu',
                        help_string='''\
Where libvirt (qemu) stores the vms config files.'''))

# The base directories where vms are stored for lxc
register(PathOption('vm.lxcs_dir', default='/var/lib/lxc',
                    help_string='''Where lxc definitions are stored.'''))
# Isos and images download handling
register(options.Option('vm.iso_url',
                        default='''\
http://cdimage.ubuntu.com/daily-live/current/''',
                        help_string='''Where an iso can be downloaded from.'''
                        ))
register(options.Option('vm.iso_name',
                        default='{vm.release}-desktop-{vm.cpu_model}.iso',
                        help_string='''The name of the iso.'''))
register(options.Option('vm.cloud_image_url',
                        default='''\
http://cloud-images.ubuntu.com/{vm.release}/current/''',
                        help_string='''\
Where a cloud image can be downloaded from.'''))
register(options.Option('vm.cloud_image_name',
                        default='''\
{vm.release}-server-cloudimg-{vm.cpu_model}-disk1.img''',
                        help_string='''The name of the cloud image.'''))
register(PathOption('vm.download_cache', default='{vm.images_dir}',
                    help_string='''Where downloads end up.'''))


# The VM classes are registered where/when needed
vm_class_registry = registries.Registry()


register(options.RegistryOption('vm.class', registry=vm_class_registry,
                                default=options.MANDATORY,
                                help_string='''\
The virtual machine technology to use.'''))
# The ubiquitous vm name
register(options.Option('vm.name', default=None, invalid='error',
                        help_string='''\
The vm name, used as a prefix for related files.'''))
# The second most important bit to define a vm: which ubuntu release ?
register(options.Option('vm.release', default=None, invalid='error',
                        help_string='''The ubuntu release name.'''))
# The third important piece to define a vm: where to store files like the
# console, the user-data and meta-data files, the ssh server keys, etc.
register(PathOption('vm.config_dir', default='{vm.vms_dir}/{vm.name}',
                    invalid='error',
                    help_string='''\
The directory where files specific to a vm are stored.

This includes the user-data and meta-data files used at install time (for
reference and easier debug) as well as the optional ssh server keys.

By default this is {vm.vms_dir}/{vm.name}. You can put it somewhere else by
redefining it as long as it ends up being unique for the vm.

{vm.vms_dir}/{vm.release}/{vm.name} may better suit your taste for example.
'''))
# The options defining the vm physical characteristics
register(options.Option('vm.ram_size', default='1024',
                        help_string='''The ram size in megabytes.'''))
register(options.Option('vm.disk_size', default='8G',
                        help_string='''The disk image size in bytes.

Optional suffixes "k" or "K" (kilobyte, 1024) "M" (megabyte, 1024k) "G"
(gigabyte, 1024M) and T (terabyte, 1024G) are supported.
'''))
register(options.Option('vm.cpus', default='1', help_string='''\
The number of cpus.'''))
register(options.Option('vm.cpu_model', default=None, invalid='error',
                        help_string='''The number of cpus.'''))
register(options.ListOption('vm.network', default='{{vm.class}.network}',
                            help_string='''\
The network setup.

The default value depends on the vm class used.
'''))

register(options.ListOption('kvm.network', default='network=default',
                            help_string='''\
The --network parameter for virt-install.

This can be specialized for each machine but the default should work in most
setups. Watch for your DHCP server exhausting its address space if you create a
lot of vms with random MAC addresses.
'''))
register(options.ListOption('lxd.network', default=None,
                            help_string='''\
The parameters for the 'lxc config device add command'.
'''))


register(options.Option('vm.meta_data', default='''\
instance-id: {vm.name}
local-hostname: {vm.name}
''',
                        invalid='error',
                        help_string='''\
The meta data for cloud-init to put in the seed.'''))

register(options.ListOption(
    'vm.ssh_opts',
    default=None,
    help_string='''\
A list of ssh options to be used when connecting to the vm via ssh.'''))

# host user related options
register(options.Option('vm.bind_home', default=False,
                        from_unicode=options.bool_from_store,
                        help_string='''\
Whether or not the user home dir should be mounted inside the vm.

The {vm.user} home directory is mounted.

This is implemented for lxc only.
'''))
register(options.Option('vm.user',
                        default_from_env=['USER'],
                        help_string='''The user for {vm.bind_home}.'''))

# Some bits that may added to user-data but are optional

register(PackageListOption('vm.packages', default=None,
                           help_string='''\
A list of package names to be installed.

If the package name starts with '@', it is interpreted as a file name
and its content, one package by line, is inserted in the list.
'''))
register(options.Option('vm.apt_proxy', default=None, invalid='error',
                        help_string='''\
A local proxy for apt to avoid repeated .deb downloads.

Example:

  vm.apt_proxy = http://192.168.0.42:3142
'''))
register(options.ListOption('vm.apt_sources', default=None,
                            help_string='''\
A list of apt sources entries to be added to the default ones.

Cloud-init already setup /etc/apt/sources.list with appropriate entries. Only
additional entries need to be specified here.
'''))
register(options.ListOption('vm.ssh_authorized_keys', default=None,
                            help_string='''\
A list of paths to public ssh keys to be authorized for the default user.'''))
register(options.ListOption('vm.ssh_keys', default=None,
                            help_string='''A list of paths to server ssh keys.

Both public and private keys can be provided. Accepted ssh key types are rsa,
dsa and ecdsa. The file names should match <type>.*[.pub].
'''))
register(options.Option('vm.update', default=False,
                        from_unicode=options.bool_from_store,
                        help_string='''Whether or not the vm should be updated.

Both apt-get update and apt-get upgrade are called if this option is set.
'''))
register(options.Option('vm.poweroff', default=True,
                        from_unicode=options.bool_from_store,
                        help_string='''\
Whether or not the vm should stop at the end of the installation.'''))
register(options.Option('vm.password', default='ubuntu', invalid='error',
                        help_string='''The ubuntu user password.'''))
register(options.Option('vm.launchpad_id',
                        help_string='''\
The launchpad login used for launchpad ssh access from the guest.'''))
# The scripts that are executed before powering off
register(PathOption('vm.root_script', default=None,
                    help_string='''\
The path to a script executed as root before powering off.

This script is executed before {vm.ubuntu_script}.
'''))
register(PathOption('vm.ubuntu_script', default=None,
                    help_string='''\
The path to a script executed as ubuntu before powering off.

This script is excuted after {vm.root_script}.
'''))
register(options.ListOption('vm.uploaded_scripts', default=None,
                            help_string='''\
A list of scripts to be uploaded to the guest.

Scripts can use config options from their vm, they will be expanded before
upload. All scripts are uploaded into {vm.uploaded_scripts.guest_dir} under
their base name.
'''))
register(options.Option('vm.uploaded_scripts.guest_dir',
                        default='~ubuntu/bin',
                        help_string='''\
Where {vm.uploaded_scripts} are uploaded on the guest.'''))
register(options.Option('vm.final_message',
                        default=None,
                        help_string='''\
The cloud-init final message, a release-specific message is used if
none is provided to address compatibility with Ubuntu precise.'''))

# timeouts: some operations may fail transiently, they have to be retried
# FIXME: two kinds of timeouts seem appropriate: a simple list to sleep times,
# a tuple defining an exponential backoff. -- vila 2015-06-25
register(options.ListOption('vm.lxc.set_ip_timeouts', default='0, 120, 10',
                            help_string='''\
A timeouts tuple to use when waiting for lxc to set an IP.

(first, up_to, retries):
- first: seconds to wait after the first attempt
- up_to: seconds after which to give up
- retries: how many attempts after the first try
'''))
register(options.ListOption('vm.lxc.ssh_setup_timeouts', default='0, 120, 10',
                            help_string='''\
A timeouts tuple to use when waiting for lxc to setup ssh.

(first, up_to, retries):
- first: seconds to wait after the first attempt
- up_to: seconds after which to give up
- retries: how many attempts after the first try
'''))
register(options.ListOption('vm.lxd.set_ip_timeouts', default='0, 120, 10',
                            help_string='''\
A timeouts tuple to use when waiting for lxd to set an IP.

(first, up_to, retries):
- first: seconds to wait after the first attempt
- up_to: seconds after which to give up
- retries: how many attempts after the first try
'''))
register(options.ListOption('vm.lxd.ssh_setup_timeouts', default='0, 120, 10',
                            help_string='''\
A timeouts tuple to use when waiting for lxd to setup ssh.

(first, up_to, retries):
- first: seconds to wait after the first attempt
- up_to: seconds after which to give up
- retries: how many attempts after the first try
'''))

# nova options
register(options.Option('vm.os.username', default_from_env=['OS_USERNAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack user name.

This is generally set via OS_USERNAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('vm.os.password', default_from_env=['OS_PASSWORD'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack password.

This is generally set via OS_PASSWORD, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('vm.os.region_name',
                        default_from_env=['OS_REGION_NAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack region name.

This is generally set via OS_REGION_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('vm.os.tenant_name',
                        default_from_env=['OS_TENANT_NAME'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack tenant name.

This is generally set via OS_TENANT_NAME, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.Option('vm.os.auth_url', default_from_env=['OS_AUTH_URL'],
                        default=options.MANDATORY,
                        help_string='''The Open Stack keystone url.

This is generally set via OS_AUTH_URL, sourced from a novarc file
(~/.novarc, ~/.canonistack/novarc).
'''))
register(options.ListOption('vm.os.flavors', default=None,
                            help_string='''\
A list of flavors for all supported clouds.

The first known one is used.
'''))
register(options.Option('vm.image', default=options.MANDATORY,
                        help_string='''The glance image to boot from.'''))
register(options.Option('vm.net_id', default=None,
                        help_string='''The network id for the vm.'''))
register(PathOption('vm.ssh_key_path', default='~/.ssh/id_rsa',
                    help_string='''The ssh key for the vm.'''))
register(options.ListOption('vm.apt_get.update.timeouts',
                            default='15.0, 90.0, 240.0',
                            help_string='''apt-get update timeouts in seconds.

When apt-get update fails on hash sum mismatches, retry after the specified
timeouts. More values mean more retries.
'''))
# FIXME: According to the help string this can (and should) be fixed
# -- vila 2015-07-20
register(options.ListOption('vm.ppas',
                            default=None,
                            help_string='''PPAs to be added to the vm.

This works around cloud-init not activating the deb-src line and not providing
a way to do so. This is intended to be fixed in uci-vms so vm.apt_sources can
be used again.
'''))
register(options.Option('vm.nova.boot_timeout', default='600',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time to boot a nova instance (in seconds).'''))
register(options.Option('vm.nova.set_ip_timeout', default='300',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time for a nova instance to get an IP (in seconds).'''))
# FIXME: 600 is a more realistic value but canonistack can be really slow and
# 3600 has been observed. There should be a better way to declare default
# values on a per cloud basis -- vila 2015-12-21
register(options.Option('vm.nova.cloud_init_timeout', default='3600',
                        from_unicode=options.float_from_store,
                        help_string='''\
Max time for cloud-init to finish (in seconds).'''))



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import io
import os
import unittest

from uciconfig import (
    options,
    registries,
)
from ucitests import (
    assertions,
    fixtures,
    scenarii,
)
from ucivms import (
    commands,
    config,
    errors,
    vms,
)
from ucivms.tests import (
    features as vms_features,
    fixtures as vms_fixtures,
)

load_tests = scenarii.load_tests_with_scenarios


class TestHelpOptions(unittest.TestCase):

    def setUp(self):
        super(TestHelpOptions, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        help_cmd = commands.Help(out=self.out, err=self.err)
        return help_cmd.parse_args(args)

    def test_defaults(self):
        ns = self.parse_args([])
        self.assertEqual([], ns.commands)

    def test_single_command(self):
        ns = self.parse_args(['setup'])
        self.assertEqual(['setup'], ns.commands)

    def test_several_commands(self):
        ns = self.parse_args(['help', 'setup', 'not-a-command'])
        self.assertEqual(['help', 'setup', 'not-a-command'], ns.commands)


class TestHelp(unittest.TestCase):

    def setUp(self):
        super(TestHelp, self).setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()

    def assertHelp(self, expected, args=None):
        if args is None:
            args = []
        help_cmd = commands.Help(out=self.out, err=self.err)
        help_cmd.parse_args(args)
        help_cmd.run()
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_help_alone(self):
        self.assertHelp('''\
Available commands:
\tconfig: Manage a virtual machine configuration.
\thelp: Describe uci-vms commands.
\tsetup: Setup a virtual machine.
\tshell: Run a command on an existing virtual machine.
\tstart: Start an existing virtual machine.
\tstatus: Display the status of a virtual machine.
\tstop: Stop an existing virtual machine.
\tteardown: Teardown a virtual machine.
''')

    def test_help_help(self):
        self.assertHelp('''\
usage: uci-vms... help [-h] [COMMAND [COMMAND ...]]

Describe uci-vms commands.

positional arguments:
  COMMAND     Display help for each command. List all command descriptions if
              none is given.

optional arguments:
  -h, --help  show this help message and exit
''',
                        ['help'])


class FakeVM(vms.VM):
    """A fake VM for tests that doesn't trigger dangerous or costly calls."""

    def __init__(self, conf, vm_states=None):
        super(FakeVM, self).__init__(conf)
        self.states = vm_states
        self.setup_called = False
        self.start_called = False
        self.shell_called = False
        self.shell_command = None
        self.shell_cmd_args = None
        self.teardown_called = False
        self.stop_called = False

    def state(self):
        return self.states.get(self.conf.get('vm.name'), None)

    def setup(self):
        self.setup_called = True

    def start(self):
        self.start_called = True

    def shell(self, command, *cmd_args):
        self.shell_called = True
        self.shell_command = command
        self.shell_cmd_args = cmd_args

    def stop(self):
        self.stop_called = True

    def teardown(self):
        self.teardown_called = True


def setup_fake_vm(test):
    vms_fixtures.isolate_from_disk(test)
    test.conf = config.VmStack('foo')
    test.conf.store._load_from_string('''
[foo]
vm.name=foo
vm.class=fake
vm.release=saucy
vm.cpu_model=amd64
''')
    test.addCleanup(test.conf.store.save_changes)
    # Register our fake vm class inside the registry option as that's where it
    # matters.
    vm_reg = options.option_registry.get('vm.class')
    fixtures.patch(test, vm_reg, 'registry', registries.Registry())
    # Only one kind of vm exists in the temporary registry
    vm_reg.registry.register('fake', FakeVM, 'Fake vm')
    # Install a way to decide if a given vm is running or not
    test.states = {}


class TestVmCommandOptions(unittest.TestCase):

    def setUp(self):
        super(TestVmCommandOptions, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        vm_cmd = commands.VmCommand(out=self.out, err=self.err)
        return vm_cmd.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)


class TestVmCommand(unittest.TestCase):

    def setUp(self):
        super(TestVmCommand, self).setUp()
        setup_fake_vm(self)

    def run_cmd(self, args):
        vm_cmd = commands.VmCommand()
        vm_cmd.parse_args(args)
        vm_cmd.run()

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.UciVmsError):
            self.run_cmd(['I-dont-exist'])

    def test_unknown_vm_class(self):
        self.conf.set('vm.class', 'I-dont-exist')
        with self.assertRaises(errors.InvalidVmClass):
            self.run_cmd(['foo'])


class TestConfigOptions(unittest.TestCase):

    def setUp(self):
        super(TestConfigOptions, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        conf_cmd = commands.Config(out=self.out, err=self.err)
        return conf_cmd.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertIs(None, ns.name)
        self.assertFalse(ns.all)
        self.assertFalse(ns.remove)

    def test_name(self):
        ns = self.parse_args(['foo', 'bar'])
        self.assertEqual('bar', ns.name)

    def assertError(self, expected):
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.err.getvalue())

    def test_too_much_args(self):
        with self.assertRaises(SystemExit) as cm:
            self.parse_args(['foo', 'bar', 'baz'])
        self.assertEqual(2, cm.exception.code)
        self.assertError('''\
usage: uci-vms... config [-h] [--remove] [--all] vm_name [name]
uci-vms... config: error: unrecognized arguments: baz
''')

    def test_remove_requires_name(self):
        with self.assertRaises(errors.UciVmsError) as cm:
            self.parse_args(['foo', '--remove'])
        self.assertEqual('--remove expects an option to remove.',
                         '{}'.format(cm.exception))

    def test_remove_all_fails(self):
        with self.assertRaises(errors.UciVmsError) as cm:
            self.parse_args(['foo', '--remove', '--all'])
        self.assertEqual('--remove and --all are mutually exclusive.',
                         '{}'.format(cm.exception))

    def test_set_only_one_option(self):
        with self.assertRaises(errors.UciVmsError) as cm:
            self.parse_args(['foo', 'bar=baz', '--all'])
        self.assertEqual('Only one option can be set.',
                         '{}'.format(cm.exception))


class TestConfig(unittest.TestCase):

    def setUp(self):
        super(TestConfig, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def run_config(self, args):
        self.vm = FakeVM(self.conf, self.states)
        cmd_config = commands.Config(_vm=self.vm, out=self.out, err=self.err)
        cmd_config.parse_args(args)
        cmd_config.run()

    def test_set_new_option(self):
        self.assertEqual(None, self.conf.get('foo'))
        self.run_config(['foo', 'foo=bar'])
        self.assertEqual('bar', self.conf.get('foo'))

    def test_set_existing_option(self):
        self.conf.set('foo', 'bar')
        self.assertEqual('bar', self.conf.get('foo'))
        self.run_config(['foo', 'foo=qux'])
        self.assertEqual('qux', self.conf.get('foo'))

    def test_remove_unknown_option(self):
        with self.assertRaises(errors.ConfigOptionNotFound):
            self.run_config(['foo', '-r', 'foo'])

    def test_remove_existing_option(self):
        self.conf.set('bar', 'baz')
        self.run_config(['foo', '-r', 'bar'])
        self.assertEqual(None, self.conf.get('bar'))

    def assertOutput(self, expected):
        assertions.assertMultiLineAlmostEqual(self, expected,
                                              self.out.getvalue())

    def test_list_option(self):
        self.conf.set('foo', 'bar')
        self.run_config(['foo', 'foo'])
        self.assertOutput('bar')

    def test_unknown_option(self):
        with self.assertRaises(errors.ConfigOptionNotFound):
            self.run_config(['foo', 'foo'])
        self.assertOutput('')

    def test_list_that_conf(self):
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/uci-vms.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = saucy
  vm.cpu_model = amd64
''')

    def test_more_options(self):
        self.conf.set('one', '1')
        self.conf.set('two', '2')
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/uci-vms.conf:
  [foo]
  vm.name = foo
  vm.class = fake
  vm.release = saucy
  vm.cpu_model = amd64
  one = 1
  two = 2
''')

    def test_options_several_sections(self):
        self.conf.store.unload()
        self.conf.store._load_from_string('''\
one = 1
two = 2
[foo]
one = 1
two = foo
[bar]
one = bar
two = foo
''')
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/uci-vms.conf:
  [foo]
  one = 1
  two = foo
  []
  one = 1
  two = 2
''')

    def test_home_config_shown_once_only(self):
        os.chdir('home')
        self.conf.store.unload()
        self.conf.store._load_from_string('''\
one = 1
[foo]
one = foo
[bar]
one = bar
''')
        self.conf.store.save()
        # Reload the stack from the current dir
        self.conf = config.VmStack('foo')
        self.run_config(['foo'])
        self.assertOutput('''\
.../home/uci-vms.conf:
  [foo]
  one = foo
  []
  one = 1
''')


class TestSetupOptions(unittest.TestCase):

    def setUp(self):
        super(TestSetupOptions, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        self.vm = FakeVM(self.conf, self.states)
        setup = commands.Setup(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_nothing(self):
        with self.assertRaises(SystemExit):
            self.parse_args([])

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertFalse(ns.download)
        self.assertFalse(ns.ssh_keygen)

    def test_download(self):
        ns = self.parse_args(['foo', '--download'])
        self.assertEqual('foo', ns.vm_name)
        self.assertTrue(ns.download)
        self.assertFalse(ns.ssh_keygen)

    def test_ssh_keygen(self):
        ns = self.parse_args(['foo', '--ssh-keygen'])
        self.assertEqual('foo', ns.vm_name)
        self.assertFalse(ns.download)
        self.assertTrue(ns.ssh_keygen)


class TestSetup(unittest.TestCase):

    def setUp(self):
        super(TestSetup, self).setUp()
        setup_fake_vm(self)

    def run_setup(self, args):
        self.vm = FakeVM(self.conf, self.states)
        setup = commands.Setup(_vm=self.vm)
        setup.parse_args(args)
        setup.run()

    def test_not_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'shut off'}
        self.run_setup(['foo'])
        self.assertTrue(self.vm.setup_called)
        self.assertTrue(self.vm.teardown_called)

    def test_while_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_setup(['foo'])
        self.assertFalse(self.vm.setup_called)
        self.assertFalse(self.vm.teardown_called)


class TestStatus(unittest.TestCase):

    def setUp(self):
        super(TestStatus, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def run_status(self, args):
        self.vm = FakeVM(self.conf, self.states)
        status = commands.Status(_vm=self.vm, out=self.out, err=self.err)
        status.parse_args(args)
        status.run()

    def test_not_setup(self):
        self.states = {'foo': 'UNKNOWN'}
        self.run_status(['foo'])
        self.assertEqual('UNKNOWN\n', self.out.getvalue())

    def test_not_running(self):
        self.states = {'foo': 'shut off'}
        self.run_status(['foo'])
        self.assertEqual('STOPPED\n', self.out.getvalue())

    def test_running(self):
        self.states = {'foo': 'running'}
        self.run_status(['foo'])
        self.assertEqual('RUNNING\n', self.out.getvalue())


class TestStatusPerVm(unittest.TestCase):

    scenarios = [(k, dict(kls=k)) for k in config.vm_class_registry.keys()]

    def setUp(self):
        super(TestStatusPerVm, self).setUp()
        # FIXME: If tests config was properly shared across all tests we could
        # get rid of the specific reference to nova below -- vila 2015-08-26
        if self.kls == 'nova':
            nova_creds = vms_features.nova_creds
            if not nova_creds.available():
                self.skipTest('{} is not available'.format(
                    nova_creds.feature_name()))
            vms_fixtures.share_nova_test_creds(self)
        elif 'lxc' in self.kls:
            # FIXME: scenario-dependent check ? Urgh. -- vila 2015-09-30
            sudo = vms_features.use_sudo_for_tests_feature
            if not sudo.available():
                self.skipTest('{} is not available'.format(
                    sudo.feature_name()))
        elif 'lxd' in self.kls:
            # FIXME: scenario-dependent check ? Urgh. -- vila 2016-01-18
            lxd = vms_features.lxd_client_feature
            if not lxd.available():
                self.skipTest('{} is not available'.format(
                    lxd.feature_name()))
        self.conf = config.VmStack('foo')
        self.conf.set('vm.name', 'foo')
        self.conf.set('vm.class', self.kls)
        self.conf.set('vm.release', 'trusty')
        self.conf.set('vm.cpu_model', 'amd64')
        self.addCleanup(self.conf.store.save_changes)

    def test_unknown_vm(self):
        vm_class = self.conf.get('vm.class')
        vm = vm_class(self.conf)
        self.assertEqual('UNKNOWN', vm.state())


class TestStart(unittest.TestCase):

    def setUp(self):
        super(TestStart, self).setUp()
        setup_fake_vm(self)

    def run_start(self, args):
        self.vm = FakeVM(self.conf, self.states)
        start = commands.Start(_vm=self.vm)
        start.parse_args(args)
        start.run()

    def test_not_running(self):
        self.states = {'foo': 'shut off'}
        self.run_start(['foo'])
        self.assertTrue(self.vm.start_called)

    def test_while_running(self):
        self.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_start(['foo'])
        self.assertFalse(self.vm.start_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_start(['I-dont-exist'])
        self.assertFalse(self.vm.start_called)


class TestShellOptions(unittest.TestCase):

    def setUp(self):
        super(TestShellOptions, self).setUp()
        setup_fake_vm(self)
        self.out = io.StringIO()
        self.err = io.StringIO()

    def parse_args(self, args):
        self.vm = FakeVM(self.conf, self.states)
        setup = commands.Shell(_vm=self.vm, out=self.out, err=self.err)
        return setup.parse_args(args)

    def test_defaults(self):
        ns = self.parse_args(['foo'])
        self.assertEqual('foo', ns.vm_name)
        self.assertIsNone(ns.command)
        self.assertEqual([], ns.args)

    def test_command_without_arguments(self):
        ns = self.parse_args(['foo', 'doit'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('doit', ns.command)
        self.assertEqual([], ns.args)

    def test_command_with_arguments(self):
        ns = self.parse_args(['foo', 'doit', '-a', 'b', 'c'])
        self.assertEqual('foo', ns.vm_name)
        self.assertEqual('doit', ns.command)
        self.assertEqual(['-a', 'b', 'c'], ns.args)


class TestShell(unittest.TestCase):

    def setUp(self):
        super(TestShell, self).setUp()
        setup_fake_vm(self)

    def run_shell(self, args):
        self.vm = FakeVM(self.conf, self.states)
        shell = commands.Shell(_vm=self.vm)
        shell.parse_args(args)
        shell.run()

    def test_not_running(self):
        self.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_shell(['foo'])
        self.assertFalse(self.vm.shell_called)
        self.assertIsNone(self.vm.shell_command)
        self.assertIsNone(self.vm.shell_cmd_args)

    def test_while_running_no_command(self):
        self.states = {'foo': 'running'}
        self.run_shell(['foo'])
        self.assertTrue(self.vm.shell_called)
        self.assertIsNone(self.vm.shell_command)
        self.assertEqual((), self.vm.shell_cmd_args)

    def test_while_running_with_command(self):
        self.states = {'foo': 'running'}
        self.run_shell(['foo', 'doit', 'a', 'b', 'c'])
        self.assertTrue(self.vm.shell_called)
        self.assertEqual('doit', self.vm.shell_command)
        self.assertEqual(('a', 'b', 'c'), self.vm.shell_cmd_args)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        with self.assertRaises(errors.VmUnknown):
            self.run_shell(['I-dont-exist'])
        self.assertFalse(self.vm.shell_called)


class TestStop(unittest.TestCase):

    def setUp(self):
        super(TestStop, self).setUp()
        setup_fake_vm(self)

    def run_stop(self):
        self.vm = FakeVM(self.conf, self.states)
        stop = commands.Stop(_vm=self.vm)
        stop.parse_args(['foo'])
        stop.run()

    def test_not_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'shut off'}
        with self.assertRaises(errors.VmNotRunning):
            self.run_stop()
        self.assertFalse(self.vm.stop_called)

    def test_while_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'running'}
        self.run_stop()
        self.assertTrue(self.vm.stop_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        self.states = {}
        with self.assertRaises(errors.VmUnknown):
            self.run_stop()
        self.assertFalse(self.vm.stop_called)


class TestTeardown(unittest.TestCase):

    def setUp(self):
        super(TestTeardown, self).setUp()
        setup_fake_vm(self)

    def run_teardown(self, args):
        self.vm = FakeVM(self.conf, self.states)
        teardown = commands.Teardown(_vm=self.vm)
        teardown.parse_args(args)
        teardown.run()

    def test_not_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'shut off'}
        self.run_teardown(['foo'])
        self.assertTrue(self.vm.teardown_called)

    def test_while_running(self):
        self.conf.set('vm.name', 'foo')
        self.states = {'foo': 'running'}
        with self.assertRaises(errors.VmRunning):
            self.run_teardown(['foo'])
        self.assertFalse(self.vm.teardown_called)

    def test_unknown(self):
        self.conf.set('vm.name', 'I-dont-exist')
        self.states = {}
        with self.assertRaises(errors.VmUnknown):
            self.run_teardown(['I-dont-exist'])
        self.assertFalse(self.vm.teardown_called)



#!/usr/bin/env python3

# This file is part of the Ubuntu Continuous Integration virtual machine tools
#
# Copyright 2014, 2015 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


import setuptools
import sys


import ucivms


def get_scripts():
    if sys.version_info < (3,):
        return ['uci-vms2']
    else:
        return ['uci-vms']


setuptools.setup(
    name='ucivms',
    version='.'.join(str(c) for c in ucivms.__version__[0:3]),
    description=('Ubuntu Continuous Integration virtual machine tools.'),
    author='Vincent Ladeuil',
    author_email='vila+qa@canonical.com',
    url='https://launchpad.net/uci-vms',
    license='GPLv3',
    install_requires=['uciconfig'],
    packages=['ucivms', 'ucivms.tests', 'ucivms.vms'],
    scripts=get_scripts(),
)



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Setup a virtual machine from a config file.

Note: Most of the operations requires root access and this script uses ``sudo``
to get them.
"""
from __future__ import unicode_literals
import io
import errno
import os
import time
try:
    import urlparse  # python2
except:
    from urllib import parse as urlparse


import yaml

from ucivms import (
    errors,
    logs,
    monitors,
    ssh,
    subprocesses,
    timeouts,
)


class CIUserData(object):
    """Maps configuration data into cloud-init user-data.

    This is a container for the data that will ultimately be encoded into a
    cloud-config-archive user-data file.
    """

    def __init__(self, conf):
        super(CIUserData, self).__init__()
        self.conf = conf
        # The objects we will populate before creating a yaml encoding as a
        # cloud-config-archive file
        self.cloud_config = {}
        self.root_hook = None
        self.ubuntu_hook = None
        self.launchpad_hook = None
        self.uploaded_scripts_hook = None

    def set(self, ud_name, conf_name=None, value=None):
        """Set a user-data option from its corresponding configuration one.

        :param ud_name: user-data key.

        :param conf_name: configuration key, If set to None, `value` should be
            provided.

        :param value: value to use if `conf_name` is None.
        """
        if value is None and conf_name is not None:
            value = self.conf.get(conf_name)
        if value:
            self.cloud_config[ud_name] = value

    def _file_content(self, path, option_name):
        full_path = os.path.expanduser(path)
        try:
            with open(full_path) as f:
                content = f.read()
        except IOError as e:
            if e.args[0] == errno.ENOENT:
                raise errors.ConfigPathNotFound(path, option_name)
            else:
                raise
        return content

    def set_list_of_paths(self, ud_name, conf_name):
        """Set a user-data option from its corresponding configuration one.

        The configuration option is a list of paths and the user-data option
        will be a list of each file content.

        :param ud_name: user-data key.

        :param conf_name: configuration key.
        """
        paths = self.conf.get(conf_name)
        if paths:
            contents = []
            for p in paths:
                contents.append(self._file_content(p, conf_name))
            self.set(ud_name, None, contents)

    def _key_from_path(self, path, option_name):
        """Infer user-data key from file name."""
        ssh_type, kind = ssh.infos_from_path(path)
        if ssh_type is None:
            raise errors.ConfigValueError(option_name, path)
        return '%s_%s' % (ssh_type, kind)

    def set_ssh_keys(self):
        """Set the server ssh keys from a list of paths.

        Provided paths should respect some coding:

        - the base name should start with the ssh type of their key (rsa, dsa,
          ecdsa),

        - base names ending with '.pub' are for public keys, the others are for
          private keys.
        """
        key_paths = self.conf.get('vm.ssh_keys')
        if key_paths:
            ssh_keys = {}
            for p in key_paths:
                key = self._key_from_path(p, 'vm.ssh_keys')
                ssh_keys[key] = self._file_content(p, 'vm.ssh_keys')
            self.set('ssh_keys', None, ssh_keys)

    def set_apt_sources(self):
        sources = self.conf.get('vm.apt_sources')
        if sources:
            apt_sources = []
            for src in sources:
                # '|' should not appear in urls nor keys so it should be safe
                # to use it as a separator.
                parts = src.split('|')
                if len(parts) == 1:
                    apt_sources.append({'source': parts[0]})
                else:
                    # For PPAs, an additional GPG key should be imported in the
                    # guest.
                    apt_sources.append({'source': parts[0], 'keyid': parts[1]})
            self.cloud_config['apt_sources'] = apt_sources

    def append_cmd(self, cmd):
        cmds = self.cloud_config.get('runcmd', [])
        cmds.append(cmd)
        self.cloud_config['runcmd'] = cmds

    def _hook_script_path(self, user):
        return '/tmp/{}-uci-vms_post_install'.format(user)

    def _hook_content(self, option_name, user, hook_path, mode='0700'):
        # FIXME: Add more tests towards properly creating a tree on the guest
        # from a tree on the host. There seems to be three kind of items worth
        # caring about here: file content (output path, owner, chmod), file
        # (input and output paths, owner, chmod) and directory (path, owner,
        # chmod). There are also some subtle traps involved about when files
        # are created across various vm generations (one vm creates a dir, a mv
        # on top of that one doesn't, but still creates a file in this dir,
        # without realizing it can fail in a fresh vm). -- vila 2013-03-10
        host_path = self.conf.get(option_name)
        if host_path is None:
            return None
        fcontent = self._file_content(host_path, option_name)
        # Expand options in the provided content so we report better errors
        expanded_content = self.conf.expand_options(fcontent)
        # The following will generate an additional newline at the end of the
        # script. I can't think of a case where it matters and it makes this
        # code more robust (and/or simpler) if the script/file *doesn't* end up
        # with a proper newline.
        # FIXME: This may be worth fixing if we provide a more generic way to
        # create a remote tree. -- vila 2013-03-10
        hook_content = '''#!/bin/sh
cat >{__guest_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
{__fcontent}
EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {__user}:{__user} {__guest_path}
chmod {__mode} {__guest_path}
'''
        return hook_content.format(__user=user, __fcontent=expanded_content,
                                   __mode=mode,
                                   __guest_path=hook_path)

    def set_boot_hook(self):
        # FIXME: Needs a test ensuring we execute as root -- vila 2013-03-07
        hook_path = self._hook_script_path('root')
        content = self._hook_content('vm.root_script', 'root', hook_path)
        if content is not None:
            self.root_hook = content
            self.append_cmd(hook_path)

    def set_ubuntu_hook(self):
        # FIXME: Needs a test ensuring we execute as ubuntu -- vila 2013-03-07
        host_path = self.conf.get('vm.ubuntu_script')
        if host_path is None:
            return
        fcontent = self._file_content(host_path, 'vm.ubuntu_script')
        if fcontent is not None:
            # Expand options in the provided content so we report better errors
            self.ubuntu_hook = self.conf.expand_options(fcontent)

    def set_launchpad_access(self):
        # FIXME: Needs a test that we can really access launchpad properly via
        # ssh. Can only be done as a real launchpad user and as such requires
        # cooperation :) I.e. Some configuration option set by the user will
        # trigger the test -- vila 2013-03-14
        lp_id = self.conf.get('vm.launchpad_id')
        if lp_id is None:
            return
        # FIXME: There is no config option to define the key name
        # (<lp_id>@uci-vms) -- vila 2014-01-16
        # Use the specified ssh key found in ~/.ssh as the private key. The
        # user is supposed to have uploaded the public one.
        local_path = os.path.join('~', '.ssh', '%s@uci-vms' % (lp_id,))
        # Force id_rsa or we'll need a .ssh/config to point to user@uci-vms
        # for .lauchpad.net.
        hook_path = '/home/ubuntu/.ssh/id_rsa'
        dir_path = os.path.dirname(hook_path)
        try:
            fcontent = self._file_content(local_path, 'vm.launchpad_id')
        except errors.ConfigPathNotFound as e:
            # Override the fmt for this specific exception to give a better
            # explanation.
            e.fmt = ('You need to create the {p} keypair and upload {p}.pub to'
                     ' launchpad.\n'
                     'See vm.launchpad_id in README.'.format(p=local_path))
            raise e
        # FIXME: ~Duplicated from _hook_content. -- vila 2013-03-10

        # FIXME: If this hook is executed before the ubuntu user is created we
        # need to chown/chmod ~ubuntu which is bad. This happens when a
        # -pristine vm is created and lead to GUI login failing because it
        # can't create any dir/file there. The fix is to only create a script
        # that will be executed via runcmd so it will run later and avoid the
        # issue. -- vila 2013-03-21
        # FIXME: Moreover, -pristine vms don't have bzr installed so this
        # cannot succeed there -- vila 2013-08-07
        hook_content = '''#!/bin/sh
mkdir -p {dir_path}
chown {user}:{user} {dir_path}
chmod {dir_mode} {dir_path}
cat >{guest_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
{fcontent}
EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {user}:{user} {guest_path}
chmod {file_mode} {guest_path}
sudo -u ubuntu bzr launchpad-login {lp_id}
'''
        self.launchpad_hook = self.conf.expand_options(
            hook_content,
            env=dict(user='ubuntu', fcontent=fcontent,
                     file_mode='0400', guest_path=hook_path,
                     dir_mode='0700', dir_path=dir_path,
                     lp_id=lp_id))

    def set_uploaded_scripts(self):
        script_paths = self.conf.get('vm.uploaded_scripts')
        if not script_paths:
            return
        hook_path = '~ubuntu/uci-vms_uploads'
        bindir = self.conf.get('vm.uploaded_scripts.guest_dir')
        out = io.StringIO()
        out.write('''#!/bin/sh
cat >{hook_path} <<'EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN'
mkdir -p {bindir}
cd {bindir}
'''.format(**locals()))
        for path in script_paths:
            fcontent = self._file_content(path, 'vm.uploaded_scripts')
            expanded = self.conf.expand_options(fcontent)
            base = os.path.basename(path)
            # FIXME: ~Duplicated from _hook_content. -- vila 2012-03-15
            out.write('''cat >{base} <<'EOF{base}'
{expanded}
EOF{base}
chmod 0755 {base}
'''.format(**locals()))

        out.write('''EOSETUPVMUNIQUECONTENTDONTBREAKFORFUN
chown {user}:{user} {hook_path}
chmod 0700 {hook_path}
'''.format(user='ubuntu', **locals()))
        self.uploaded_scripts_hook = out.getvalue()
        self.append_cmd(['su', '-l', '-c', hook_path, 'ubuntu'])

    def set_poweroff(self):
        # We want to shutdown properly after installing. This is safe to set
        # here as subsequent boots will ignore this setting, letting us use the
        # vm ;)
        if self.conf.get('vm.poweroff'):
            if self.conf.get('vm.release') in ('precise', 'quantal'):
                # Curse cloud-init lack of compatibility
                self.append_cmd('halt')
            else:
                self.set('power_state', None, {'mode': 'poweroff'})

    def populate(self):
        # Common and non-configurable options
        self.set('manage_etc_hosts', None, True)
        self.set('chpasswd', None, dict(expire=False))
        # Configurable options
        msg = self.conf.get('vm.final_message')
        if msg is None:
            if self.conf.get('vm.release') == 'precise':
                # Curse cloud-init lack of compatibility
                msg = 'uci-vms finished installing in $UPTIME seconds.'
            else:
                msg = 'uci-vms finished installing in ${uptime} seconds.'
        self.set('final_message', None, msg)
        self.set('password', 'vm.password')
        self.set_list_of_paths('ssh_authorized_keys', 'vm.ssh_authorized_keys')
        self.set_ssh_keys()
        self.set('apt_proxy', 'vm.apt_proxy')
        # Both user-data keys are set from the same config key, we don't
        # provide a finer access.
        self.set('apt_update', 'vm.update')
        self.set('apt_upgrade', 'vm.update')
        self.set_apt_sources()
        self.set('packages', 'vm.packages')
        self.set_launchpad_access()
        # uploaded scripts
        self.set_uploaded_scripts()
        # The commands executed before powering off
        self.set_boot_hook()
        self.set_ubuntu_hook()
        # This must be called last so previous commands (for precise and
        # quantal) can be executed before powering off
        self.set_poweroff()

    def add_boot_hook(self, parts, hook):
        """Add a cloud-boothook.

        Note that this hook is excuted immediately, run at every boot at an
        early stage where parts of the system may not be available (including
        the ubuntu user).

        :params parts: The existing parts of the cloud-config-archive.

        :params hook: The hook to add in textual form.
        """
        if hook is not None:
            parts.append({'content': '#cloud-boothook\n' + hook})

    def add_user_data_script(self, parts, hook):
        """Add a User-Data script.

        The received script is executed very late in the boot sequence
        (rc.local-like).

        :params parts: The existing parts of the cloud-config-archive.

        :params hook: The hook to add in textual form. It MUST start with a
            shebang line '#!'.
        """
        if hook is not None:
            parts.append({'content': hook})

    def dump(self):
        def yaml_dumps(obj):
            return yaml.safe_dump(obj)

        parts = [{'content':
                  '#cloud-config\n' + yaml_dumps(self.cloud_config)}]
        self.add_boot_hook(parts, self.root_hook)
        self.add_user_data_script(parts, self.launchpad_hook)
        self.add_user_data_script(parts, self.uploaded_scripts_hook)
        self.add_user_data_script(parts, self.ubuntu_hook)
        # Wrap the lot into a cloud config archive
        return '#cloud-config-archive\n' + yaml_dumps(parts)


class VM(object):
    """A virtual machine relying on cloud-init to customize installation."""

    def __init__(self, conf):
        self.conf = conf
        # Seed files
        self._meta_data_path = None
        self._user_data_path = None
        self.ci_user_data = None

    def download(self, force=False):
        raise NotImplementedError(self.download)

    def setup(self):
        raise NotImplementedError(self.setup)

    def start(self):
        raise NotImplementedError(self.start)

    def stop(self):
        raise NotImplementedError(self.stop)

    def teardown(self):
        raise NotImplementedError(self.teardown)

    # MISSINGTEST
    def shell(self, command, *args):
        ssh_cmd = self.get_ssh_command(command, *args)
        retcode = subprocesses.raw_run(ssh_cmd)
        return retcode

    def shell_captured(self, command, *args):
        ssh_cmd = self.get_ssh_command(command, *args)
        retcode, out, err = subprocesses.run(ssh_cmd)
        return retcode, out, err

    def _download_in_cache(self, source_url, name, force=False):
        """Download ``name`` from ``source_url`` in ``vm.download_cache``.

        :param source_url: The url where the file to download is located

        :param name: The name of the file to download (also used as the name
            for the downloaded file).

        :param force: Remove the file from the cache if present.

        :return: False if the file is in the download cache, True if a download
            occurred.
        """
        source = urlparse.urljoin(source_url, name)
        download_dir = self.conf.get('vm.download_cache')
        if not os.path.exists(download_dir):
            raise errors.ConfigValueError('vm.download_cache', download_dir)
        target = os.path.join(download_dir, name)
        # FIXME: By default the download dir may be under root control, but if
        # a user chose to use a different one under his own control, it would
        # be nice to not require sudo usage. -- vila 2013-02-06
        if force:
            subprocesses.run(['sudo', 'rm', '-f', target])
        if not os.path.exists(target):
            # FIXME: We do ask for a progress bar but it's not displayed
            # (subprocesses.run capture both stdout and stderr) ! At least
            # while used interactively, it should. -- vila 2013-02-06
            subprocesses.run(['sudo', 'wget', '--progress=dot:mega', '-O',
                              target, source])
            return True
        else:
            return False

    def ensure_dir(self, path):
        try:
            os.makedirs(path)
        except OSError as e:
            # python2 doesn't provide the exist_ok=True to makedirs
            if e.errno != errno.EEXIST:
                raise

    def config_dir_path(self):
        return self.conf.get('vm.config_dir')

    def console_path(self):
        # The console is used to record 'setup' and 'start' outputs
        return os.path.join(self.config_dir_path(), 'console')

    def empty_console(self):
        # The console file is created during setup. As such it's created by
        # root. Care should be taken to ensure the file receives proper chmod
        # bits (or only root will be able to read it) and that it's reset at
        # the right time ('setup' is a good time as it means the console will
        # contain the creation of the instance and all its 'start' uses).
        self.ensure_dir(self.config_dir_path())
        console_path = self.console_path()
        # The console may end up being owned by root, delete it first
        try:
            os.remove(console_path)
        except OSError as e:  # python2 doesn't provide FileNotFoundError
            # It doesn't exist ? Job done.
            if e.errno != errno.ENOENT:
                raise
        # Create the file so we get access to it (otherwise it may be owned by
        # root).
        open(console_path, 'w').close()

    def ssh_keygen(self, force=False):
        self.ensure_dir(self.config_dir_path())
        keys = self.conf.get('vm.ssh_keys')
        for key in keys:
            if force or not os.path.exists(key):
                ssh.keygen(key, self.conf.get('vm.name'))

    def create_meta_data(self):
        self.ensure_dir(self.config_dir_path())
        self._meta_data_path = os.path.join(self.config_dir_path(),
                                            'meta-data')
        with open(self._meta_data_path, 'w') as f:
            f.write(self.conf.get('vm.meta_data'))

    def create_user_data(self):
        ci_user_data = CIUserData(self.conf)
        ci_user_data.populate()
        self.ensure_dir(self.config_dir_path())
        self._user_data_path = os.path.join(self.config_dir_path(),
                                            'user-data')
        with open(self._user_data_path, 'w') as f:
            f.write(ci_user_data.dump())
        self.ci_user_data = ci_user_data

    def iface_path(self, iface):
        return os.path.join(self.config_dir_path(),
                            'interface.{}'.format(iface))

    def create_iface_file(self, iface, ip, mask, mac):
        iface_path = self.iface_path(iface)
        self.ensure_dir(self.config_dir_path())
        with open(iface_path, 'w') as f:
            f.write(' '.join((ip, mask, mac)) + '\n')

    def capture_interface(self, line, iface_re):
        """Capture the required interface if it appears in 'line'.

        :param line: A line from the vm console.

        :param iface_re: The regexp matching a cloud-init output revealing an
            interface.

        :return: True if the interface was captured, False otherwise.
        """
        match = iface_re.match(line)
        if match is not None:
            ip, mask, mac = match.groups()
            self.create_iface_file(iface_re.iface, ip, mask, mac)
            return True
        return False

    def get_interface(self, iface):
        try:
            with open(self.iface_path(iface)) as f:
                content = f.read()
                ip, mask, mac = content.split()
                return ip, mask, mac
        except IOError as e:
            # python2 does not provide FileNotFoundError
            if e.errno == errno.ENOENT:
                # Unknown interface
                return None
            else:
                raise

    def scan_console_during_setup(self, console_size, cmd):
        """Scan the console output until the end of the setup.

        We add a specific part for cloud-init to ensure we properly detect
        the end of the run.

        :param console_size: The size of the console file before 'install' is
            run.

        :param cmd: The setup command (used for error display).
        """
        console = monitors.TailMonitor(self.console_path(), console_size)
        iface = 'eth0'
        iface_re = logs.InterfaceRegexp(iface)
        try:
            for line in console.scan():
                # FIXME: We need some way to activate this dynamically (conf
                # var defaulting to environment variable OR cmdline parameter ?
                # -- vila 2013-02-11

                # print('read: [{}]'.format(line))  # so useful for debug...
                self.capture_interface(line, iface_re)
                # FIXME: This bare 'pass' smells. Something is wrong in the
                # design. -- vila 2014-01-18
                pass
        except (errors.ConsoleEOFError, errors.CloudInitError):
            # FIXME: No test covers this path -- vila 2013-02-15
            err_lines = ['Suspicious line from cloud-init.\n',
                         '\t' + console.lines[-1],
                         'Check the configuration:\n']
            with open(self._meta_data_path) as f:
                err_lines.append('meta-data content:\n')
                err_lines.extend(f.readlines())
            with open(self._user_data_path) as f:
                err_lines.append('user-data content:\n')
                err_lines.extend(f.readlines())
            raise errors.CommandError(cmd, console.proc.returncode,
                                      '\n'.join(console.lines),
                                      ''.join(err_lines))

    # MISSINGTEST -- vila 2014-01-18
    def scan_console_during_start(self, console_size, cmd):
        """Scan the console output while the instance starts.

        This is used to capture the network addresses displayed by cloud-init.

        :param console_size: The size of the console file before 'start' is
            run.

        :param cmd: The start command (used for error display).
        """
        console = monitors.TailMonitor(self.console_path(), console_size)
        # FIXME: Don't use a hard-coded value, systemd will change the naming
        # at some point in the future (ens3 has been seen on wily but was
        # promptly reverted) -- vila 2015-11-02
        iface = 'eth0'
        iface_re = logs.InterfaceRegexp(iface)
        try:
            for line in console.scan():
                if self.capture_interface(line, iface_re):
                    # We're done, no need to scan anymore (and work around the
                    # fact that 'scan' will otherwise terminate when the vm is
                    # stopped ;)
                    # FIXME: This fails to exit the loop when no IP can be
                    # provided (seen while investigating
                    # https://bugs.launchpad.net/bugs/1465196)
                    # -- vila 2015-11-01
                    return
        except errors.ConsoleEOFError:
            # We're done
            pass

    def get_ssh_command(self, command, *args, **kwargs):
        ip = kwargs.pop('ip', None)
        user = 'ubuntu'
        # Unless the home is bound
        if self.conf.get('vm.bind_home'):
            user = self.conf.get('vm.user')
        if ip is None:
            # FIXME: Don't use a hard-code value, systemd will change the
            # naming at some point in the future (ens3 has been seen on wily
            # but was promptly reverted) -- vila 2015-11-02
            iface = 'eth0'
            # FIXME: This can fail with "TypeError: 'NoneType' object is not
            # iterable" which can occur when lxc can't acquire an IP from DHCP
            # (This shouldn't be reachable anymore as the bug was earlier in
            # wait_for_ip not failing when no ip can be discovered.), or when
            # kvm can't acquire an IP at all.  -- vila 2015-10-24
            ip, _, _ = self.get_interface(iface)
        cmd = ['ssh']
        ssh_opts = self.conf.get('vm.ssh_opts')
        if ssh_opts:
            cmd += ssh_opts
        cmd += ['{}@{}'.format(user, ip)]
        if command is not None:
            cmd += [command]
            if args:
                cmd += args
        return cmd


#  From man virsh(1):
#            STATES
#
#            The State field lists 8 states for a domain, and which ones the
#            current domain is in.
#
#            running
#                The domain is currently running on a CPU
#
#            idle
#                The domain is idle, and not running or runnable.  This can be
#                caused because the domain is waiting on IO (a traditional wait
#                state) or has gone to sleep because there was nothing else for
#                it to do.
#
#            paused
#                The domain has been paused, usually occurring through the
#                administrator running virsh suspend.  When in a paused state
#                the domain will still consume allocated resources like memory,
#                but will not be eligible for scheduling by the hypervisor.
#
#            shutdown
#                The domain is in the process of shutting down, i.e. the guest
#                operating system has been notified and should be in the
#                process of stopping its operations gracefully.
#
#            shut off
#                The domain is not running.  Usually this indicates the domain
#                has been shut down completely, or has not been started.
#
#            crashed
#                The domain has crashed, which is always a violent ending.
#                Usually this state can only occur if the domain has been
#                configured not to restart on crash.
#
#            dying
#                The domain is in process of dying, but hasn't completely
#                shutdown or crashed.
#
#            pmsuspended
#                The domain has been suspended by guest power management, e.g.
#                entered into s3 state.
def kvm_states(source=None):
    """A dict of states for kvms indexed by name.

    :param source: A list of lines as produced by virsh list --all without
        decorations (header/footer).
    """
    states = {}
    if source is None:
        try:
            retcode, out, err = subprocesses.run(['virsh', 'list', '--all'])
        except OSError as e:
            if e.errno == errno.ENOENT:
                # If we can't execute virsh, there is no vms
                return states
            else:
                raise
        # Get rid of header/footer
        source = out.splitlines()[2:-1]
    for line in source:
        caret_or_id, name, state = line.split(None, 2)
        states[name] = state
    return states


class Kvm(VM):

    def __init__(self, conf):
        super(Kvm, self).__init__(conf)
        # Disk paths
        self._disk_image_path = None
        self._seed_path = None

    def state(self):
        states = kvm_states()
        try:
            state = states[self.conf.get('vm.name')]
        except KeyError:
            state = 'UNKNOWN'
        return state

    def download_iso(self, force=False):
        """Download the iso to install the vm.

        :return: False if the iso is in the download cache, True if a download
            occurred.
        """
        return self._download_in_cache(self.conf.get('vm.iso_url'),
                                       self.conf.get('vm.iso_name'),
                                       force=force)

    def download_cloud_image(self, force=False):
        """Download the cloud image to install the vm.

        :return: False if the image is in the download cache, True if a
            download occurred.
        """
        return self._download_in_cache(self.conf.get('vm.cloud_image_url'),
                                       self.conf.get('vm.cloud_image_name'),
                                       force=force)

    def download(self, force=False):
        return self.download_cloud_image(force)

    def create_seed_image(self):
        if self._meta_data_path is None:
            self.create_meta_data()
        if self._user_data_path is None:
            self.create_user_data()
        images_dir = self.conf.get('vm.images_dir')
        seed_path = os.path.join(
            images_dir, self.conf.expand_options('{vm.name}.seed'))
        subprocesses.run(
            # We create the seed in the ``vm.images_dir`` directory, so
            # ``sudo`` is required
            ['sudo',
             'genisoimage', '-output', seed_path,
             # cloud-init relies on the volid to discover its data
             '-volid', 'cidata',
             '-joliet', '-rock', '-input-charset', 'default',
             '-graft-points',
             'user-data=%s' % (self._user_data_path,),
             'meta-data=%s' % (self._meta_data_path,),
             ])
        self._seed_path = seed_path

    def create_disk_image(self):
        if self.conf.get('vm.backing') is None:
            self.create_disk_image_from_cloud_image()
        else:
            self.create_disk_image_from_backing()

    def create_disk_image_from_cloud_image(self):
        """Create a disk image from a cloud one."""
        cloud_image_path = os.path.join(
            self.conf.get('vm.download_cache'),
            self.conf.get('vm.cloud_image_name'))
        disk_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.name}.qcow2'))
        subprocesses.run(
            ['sudo', 'qemu-img', 'convert',
             '-O', 'qcow2', cloud_image_path, disk_image_path])
        subprocesses.run(
            ['sudo', 'qemu-img', 'resize',
             disk_image_path, self.conf.get('vm.disk_size')])
        self._disk_image_path = disk_image_path

    def create_disk_image_from_backing(self):
        """Create a disk image backed by an existing one."""
        backing_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.backing}'))
        disk_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.name}.qcow2'))
        subprocesses.run(
            ['sudo', 'qemu-img', 'create', '-f', 'qcow2',
             '-b', backing_image_path, disk_image_path])
        subprocesses.run(
            ['sudo', 'qemu-img', 'resize',
             disk_image_path, self.conf.get('vm.disk_size')])
        self._disk_image_path = disk_image_path

    def scan_console_during_setup(self, console_size, cmd):
        """See Vm.scan_console_during_setup."""
        # The console is re-created by virt-install (even if we created it
        # before) which requires sudo but creates the file 0600 for
        # libvirt-qemu. We give read access to all otherwise 'tail -f' requires
        # sudo and can't be killed anymore.
        subprocesses.run(['sudo', 'chmod', '0644', self.console_path()])
        # While `virt-install` is running, let's connect to the console
        super(Kvm, self).scan_console_during_setup(console_size, cmd)

    def setup(self):
        # Create a kvm, relying on cloud-init to customize the base image.
        #
        # There are two processes involvded here:
        # - virt-install creates the vm and boots it.
        # - progress is monitored via the console to detect cloud-final.
        #
        # Once cloud-init has finished, the vm can be powered off.

        # FIXME: If the install doesn't finish after $time, emit a warning and
        # terminate self.install_proc.
        # FIXME: If we can't connect to the console, emit a warning and
        # terminate console and self.install_proc.
        # FIXME: If we don't receive anything on the console after $time2, emit
        # a warning and terminate console and self.install_proc.
        # -- vila 2013-02-07
        if self._seed_path is None:
            self.create_seed_image()
        if self._disk_image_path is None:
            self.create_disk_image()
        self.empty_console()
        virt_install = [
            'sudo', 'virt-install',
            # To ensure we're not bitten again by http://pad.lv/1157272 where
            # virt-install wrongly detect virtualbox. -- vila 2013-03-20
            '--connect', 'qemu:///system',
            # Without --noautoconsole, virt-install will relay the console,
            # that's not appropriate for our needs so we'll connect later
            # ourselves
            '--noautoconsole',
            # We define the console as a file so we can monitor the install
            # via 'tail -f'
            '--serial', 'file,path={}'.format(self.console_path()),
            '--network', ','.join(self.conf.get('vm.network')),
            # Anticipate that we'll need a graphic card defined
            '--graphics', 'spice',
            '--name', self.conf.get('vm.name'),
            '--ram', self.conf.get('vm.ram_size'),
            '--vcpus', self.conf.get('vm.cpus'),
            '--disk', 'path=%s,format=qcow2' % (self._disk_image_path,),
            '--disk', 'path=%s' % (self._seed_path,),
            # We just boot, cloud-init will handle the installs we need
            '--boot', 'hd', '--hvm',
        ]
        console_size = monitors.actual_file_size(self.console_path())
        subprocesses.run(virt_install)
        self.scan_console_during_setup(console_size, virt_install)
        # We've seen the console signaling halt, but the vm will need a bit
        # more time to get there so we help it a bit.
        if self.conf.get('vm.release') in ('precise', 'quantal'):
            # cloud-init doesn't implement power_state until raring and need a
            # gentle nudge.
            self.stop()
        while True:
            state = self.state()
            # We expect the vm's state to be 'in shutdown' but in some rare
            # occasions we may catch 'running' before getting 'in shutdown'.
            if state in ('in shutdown', 'running'):
                # Ok, querying the state takes time, this regulates the polling
                # enough that we don't need to sleep.
                continue
            elif state == 'shut off':
                # Good, we're done
                break
            # FIXME: No idea on how to test the following. Manually tested by
            # altering the expected state above and running 'selftest.py -v'
            # which errors out for test_setup_with_seed and
            # test_setup_backing. Also reproduced when 'running' wasn't
            # expected before 'in shutdown' -- vila 2013-02-19
            # Unexpected state reached, bad.
            raise errors.UciVmsError(
                'Something went wrong during {name} install\n'
                'The vm ended in state: {state}\n'
                'Check the console at {path}',
                name=self.conf.get('vm.name'), state=state,
                path=self.console_path())

    def start(self):
        start_cmd = ['sudo', 'virsh', 'start', self.conf.get('vm.name')]
        console_size = monitors.actual_file_size(self.console_path())
        if console_size is None:
            # FIXME: This is needed because the following can fail (real life):
            # 'rm -f ~/vms/kvm1/console ; ./uci-vms start kvm1' because
            # virsh start creates the console file as root if it doesn't exist
            self.empty_console()
            console_size = 0
        proc = subprocesses.run(start_cmd)
        self.scan_console_during_start(console_size, start_cmd)
        # FIXME: There is a race here, we need to wait for ssh to become
        # available -- vila 2015-06-18
        return proc

    def stop(self):
        # FIXME: This is async and as such needs to scan the console to return
        # only when the vm is really stopped(but we don't have live console
        # during 'virsh destroy') (or poll status until it's STOPPED, but urgh
        # polling) -- vila 2015-11-01
        return subprocesses.run(
            ['sudo', 'virsh', 'destroy', self.conf.get('vm.name')])

    def teardown(self):
        return subprocesses.run(
            ['sudo', 'virsh', 'undefine', self.conf.get('vm.name'),
             '--remove-all-storage'])


# FIXME: This works for trusty but requires a complete rewrite and tests for
# older and newer releases -- vila 2015-02-03
def lxc_info(vm_name, source=None):
    """Scan state info from the lxc-info output.

    :param vm_name: The vm we want to query about.

    :param source: A list of lines as produced by lxc-info -n vm.name.

    :returns: A dict with keys 'state' and 'pid'.
    """
    if source is None:
        try:
            retcode, out, err = subprocesses.run(
                ['sudo', 'lxc-info', '-n', vm_name])
        except errors.CommandError as e:
            # The vm doesn't exist
            if e.err.startswith("{} doesn't exist".format(vm_name)):
                return dict(state='UNKNOWN', pid='-1')
            # Re-raise otherwise
            raise
        source = out.splitlines()
    state_line = source[0]
    cur = 0
    while not state_line.startswith('State:'):
        cur += 1
        state_line = source[cur]
    if len(source) > 1 + cur:
        pid_line = source[1 + cur]
        if pid_line.startswith('Name:'):
            pid_line = 'pid: -1'
    else:
        # Recent lxc versions stopped outputting the fake pid, let's do that
        # ourselves
        pid_line = 'pid: -1'
    _, state = state_line.split(None, 1)
    _, pid = pid_line.split(None, 1)
    return dict(state=state, pid=pid)


class Lxc(VM):

    def __init__(self, conf):
        super(Lxc, self).__init__(conf)
        self._fstab_path = None
        self._lxc_base_path = None
        self._lxc_seed_path = None
        self._clear_cache = False

    def state(self):
        # man lxc(7) defines the possible states as: STOPPED, STARTING,
        # RUNNING, ABORTING, STOPPING. We add UNKNOWN.
        info = lxc_info(self.conf.get('vm.name'))
        return info['state']

    def download(self, force=False):
        self._clear_cache = force
        return True

    def create_seed_files(self):
        if self._meta_data_path is None:
            self.create_meta_data()
        if self._user_data_path is None:
            self.create_user_data()
        self._lxc_base_path = os.path.join(self.conf.get('vm.lxcs_dir'),
                                           self.conf.get('vm.name'))
        self._lxc_seed_path = os.path.join(
            self._lxc_base_path, 'rootfs/var/lib/cloud/seed/nocloud-net')

    def copy_seed_files(self):
        # FIXME: Error checking ? -- vila 2015-06-23
        mkdir_seed = ['sudo', 'mkdir', '-p', self._lxc_seed_path]
        _, out, err = subprocesses.run(mkdir_seed)
        cp_seeds = ['sudo', 'cp', self._meta_data_path, self._user_data_path,
                    self._lxc_seed_path]
        _, out, err = subprocesses.run(cp_seeds)

    def setup(self):
        '''Create an lxc, relying on cloud-init to customize the base image.

        There are two processes involvded here:
        - lxc-create creates the vm.
        - progress is monitored via the console to detect cloud-final.

        Once cloud-init has finished, the vm can be powered off.
        '''
        # FIXME: use python3-lxc ? Problematic for python2 which is still
        # needed to support nova O_o -- vila 2015-12-15

        # FIXME: If the install doesn't finish after $time, emit a warning and
        # terminate self.install_proc.
        # FIXME: If we can't connect to the console, emit a warning and
        # terminate console and self.install_proc.
        # FIXME: If we don't receive anything on the console after $time2, emit
        # a warning and terminate console and self.install_proc.
        # -- vila 2013-02-07
        if self._fstab_path is None:
            self.create_seed_files()
        self.empty_console()
        # FIXME: Some feedback would be nice during lxc creation, not sure
        # about which errors to expect there either -- vila 2013-08-07
        lxc_create = ['sudo', 'lxc-create',
                      '-n', self.conf.get('vm.name'),
                      '-t', 'ubuntu',
                      '--',
                      '-r', self.conf.get('vm.release'),
                      '-a', self.conf.get('vm.cpu_model'),
                      '--packages', 'cloud-init,ssh',
                      ]
        if self._clear_cache:
            lxc_create.append('-F')
        if self.conf.get('vm.bind_home'):
            lxc_create.extend(['--bindhome', self.conf.get('vm.user')])
        _, out, err = subprocesses.run(lxc_create)
        self.copy_seed_files()
        # FIXME: useful for debug, needs a config option -- vila 2014-01-17
        # print('lxc-create, \n\tout: {}\n\t{}'.format(out, err))
        # Now we add the cloud-init data seed and do lxc-start to trigger all
        # our customizations monitoring the lxc-start output from the host.
        lxc_start = ['sudo', 'lxc-start',
                     '-n', self.conf.get('vm.name'),
                     '--console-log', self.console_path(),
                     # Daemonize or: 1) it fails with a spurious return code,
                     # 2) We can't monitor the logfile
                     '-d',
                     ]
        # FIXME: useful for debug, needs a config option -- vila 2014-01-17
        # print 'cmd: %s' % (' '.join(lxc_start),)
        console_size = monitors.actual_file_size(self.console_path())
        ret, out, err = subprocesses.run(lxc_start)
        self.scan_console_during_setup(console_size, lxc_start)
        return ret, out, err

    # MISSINGTEST
    def wait_for_ip(self):
        ip = None
        first, up_to, retries = self.conf.get('vm.lxc.set_ip_timeouts')
        first = float(first)
        up_to = float(up_to)
        retries = int(retries)
        get_ip_cmd = ['sudo', 'lxc-info', '-n', self.conf.get('vm.name'), '-i']
        exc = None
        for sleep in timeouts.ExponentialBackoff(first, up_to, retries):
            try:
                ret, out, err = subprocesses.run(get_ip_cmd)
                if out:
                    ip = out.split()[1]
                    # FIXME: Assume eth0 until we can do better
                    # -- vila 2015-06-17
                    self.create_iface_file('eth0', ip, 'unknown', 'unknown')
                    # Success, clear the exception
                    exc = None
                    break
            except errors.CommandError as e:
                exc = e
                # FIXME: logging + metric  -- vila 2015-06-25
                pass
            time.sleep(sleep)
        if exc is not None:
            # Re-raise the last seen exception
            raise exc
        if ip is None:
            raise errors.UciVmsError('Lxc {} never provided an IP'.format(
                self.conf.get('vm.name')))

    # MISSINGTEST
    def wait_for_ssh(self):
        first, up_to, retries = self.conf.get('vm.lxc.ssh_setup_timeouts')
        first = float(first)
        up_to = float(up_to)
        retries = int(retries)
        exc = None
        for attempt, sleep in enumerate(timeouts.ExponentialBackoff(
                first, up_to, retries), start=1):
            try:
                ret, out, err = self.shell_captured('whoami')
                # Success, clear the exception
                exc = None
                break
            # FIXME: There may be better feedback to provide here if (for
            # example, the IP is lost abruptly or the vm is killed for reasons)
            # -- vila 2015-10-24
            except errors.CommandError as e:
                exc = e
                # FIXME: logging + metric  -- vila 2015-06-25
                pass
            time.sleep(sleep)
        if exc is not None:
            # Re-raise the last seen exception
            raise exc

    def start(self):
        start_cmd = ['sudo', 'lxc-start', '-n', self.conf.get('vm.name'),
                     '--console-log', self.console_path(),
                     # Starts in daemon mode or we end up with an interactive
                     # session
                     '-d',
                     ]
        console_size = monitors.actual_file_size(self.console_path())
        if console_size is None:
            # FIXME: This is needed because the following can fail (real life):
            # 'rm -f ~/vms/lxc1/console ; ./uci-vms start lxc1' because
            # lxc-start creates the console file as root if it doesn't exist
            self.empty_console()
            console_size = 0
        subprocesses.run(start_cmd)
        self.scan_console_during_start(console_size, start_cmd)
        self.wait_for_ip()
        self.wait_for_ssh()

    def stop(self):
        return subprocesses.run(
            ['sudo', 'lxc-stop', '-n', self.conf.get('vm.name')])

    def teardown(self):
        try:
            return subprocesses.run(
                ['sudo', 'lxc-destroy', '-n', self.conf.get('vm.name')])
        except errors.CommandError as e:
            # FIXME: No test -- vila 2013-08-08
            not_exist = 'does not exist\n'
            not_defined = 'is not defined\n'
            if (e.err.endswith(not_exist) or e.err.endswith(not_defined)):
                # lxc-destroy does not always make a distinction between a
                # stopped vm and a non-existing one.
                pass
            else:
                raise


class EphemeralLxc(Lxc):

    def download(self, force=False):
        raise NotImplementedError(self.download)

    def setup(self):
        raise NotImplementedError(self.setup)

    def teardown(self):
        raise NotImplementedError(self.teardown)

    def start(self):
        start_cmd = ['sudo', 'lxc-start-ephemeral',
                     '-n', self.conf.get('vm.name'),
                     '-o', self.conf.get('vm.backing'),
                     # Starts in daemon mode or we end up with an interactive
                     # session
                     '-d']
        subprocesses.run(start_cmd)
        self.wait_for_ip()
        self.wait_for_ssh()



# This file is part of Ubuntu Continuous Integration virtual machine tools.
#
# Copyright 2014, 2015, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals
import argparse
import re
import sys


from uciconfig import (
    errors as config_errors,
    registries,
)
from ucivms import (
    config,
    errors,
    vms,
)
from ucivms.vms import lxd


# The available vm backends are registered here to ensure the config registry
# is loaded before we try to use it.
config.vm_class_registry.register('kvm', vms.Kvm,
                                  'Kernel-based virtual machine')
config.vm_class_registry.register('lxc', vms.Lxc,
                                  'Linux container virtual machine')
config.vm_class_registry.register('ephemeral-lxc', vms.EphemeralLxc,
                                  'Linux container ephemeral virtual machine')
config.vm_class_registry.register('lxd', lxd.Lxd,
                                  'Lxd virtual machine')
if sys.version_info < (3,):
    try:
        from ucivms.vms import nova
    except ImportError:
        # No novaclient, no nova vms.
        pass
    else:
        # Only register nova if novaclient is available
        config.vm_class_registry.register(
            'nova', nova.NovaServer, 'Openstack Nova instance')


class ArgParser(argparse.ArgumentParser):
    """A parser for the uci-vms script."""

    def __init__(self, name, description):
        if sys.version_info < (3,):
            script_name = 'uci-vms2'
        else:
            script_name = 'uci-vms'
        super(ArgParser, self).__init__(
            prog='{} {}'.format(script_name, name),
            description=description)

    def parse_args(self, args=None, out=None, err=None):
        """Parse arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig

    def parse_known_args(self, args=None, out=None, err=None):
        """Parse known arguments, overridding stdout/stderr if provided.

        Overridding stdout/stderr is provided for tests.

        :params args: Defaults to sys.argv[1:].

        :param out: Defaults to sys.stdout.

        :param err: Defaults to sys.stderr.
        """
        out_orig = sys.stdout
        err_orig = sys.stderr
        try:
            if out is not None:
                sys.stdout = out
            if err is not None:
                sys.stderr = err
            return super(ArgParser, self).parse_known_args(args)
        finally:
            sys.stdout = out_orig
            sys.stderr = err_orig


class CommandRegistry(registries.Registry):
    """A registry specialized for commands."""

    def register(self, cmd):
        super(CommandRegistry, self).register(
            cmd.name, cmd, help_string=cmd.description)


# All commands are registered here, defining what run() supports
command_registry = CommandRegistry()


class Command(object):

    name = 'Must be set by daughter classes'
    description = 'Must be set by daughter classes'

    def __init__(self, out=None, err=None, _vm=None):
        """Command constructor.

        :param out: A stream for command output.

        :param err: A stream for command errors.

        :param _vm: An existing vms.VM instance. For tests only.
        """
        self.vm = _vm
        if out is None:
            out = sys.stdout
        if err is None:
            err = sys.stderr
        self.out = out
        self.err = err
        self.parser = ArgParser(self.name, self.description)

    def parse_args(self, args):
        self.options = self.parser.parse_args(args, self.out, self.err)
        return self.options


class Help(Command):

    name = 'help'
    description = 'Describe uci-vms commands.'

    def __init__(self, **kwargs):
        super(Help, self).__init__(**kwargs)
        self.parser.add_argument(
            'commands', metavar='COMMAND', nargs='*',
            help='Display help for each command.'
            ' List all command descriptions if none is given.')

    def run(self):
        if not self.options.commands:
            self.out.write('Available commands:\n')
            for cmd_name in command_registry.keys():
                cmd_class = command_registry.get(cmd_name)
                self.out.write('\t{}: {}\n'.format(cmd_class.name,
                                                   cmd_class.description))
            return
        for cmd_name in self.options.commands:
            try:
                cmd_class = command_registry.get(cmd_name)
            except KeyError:
                cmd_class = None
            if cmd_class:
                cmd = cmd_class()
                cmd.parser.print_help(self.out)
            else:
                msg = '"{}" is not a known command\n'.format(cmd_name)
                self.err.write(msg)
        return 0


command_registry.register(Help)


class VmCommand(Command):
    """A command applying to a given vm.

    This is not a command by itself but a base class for concrete commands.
    """

    def __init__(self, **kwargs):
        super(VmCommand, self).__init__(**kwargs)
        self.parser.add_argument(
            'vm_name',
            help='Virtual machine section in the configuration file.')

    def parse_args(self, args):
        super(VmCommand, self).parse_args(args)
        self.parse_vm_arg()
        return self.options

    def parse_vm_arg(self):
        self.vm_name = self.options.vm_name
        if self.vm is None:
            conf = config.VmStack(self.vm_name)
            # Define vm.name by adding it to the cmdline config store
            conf.cmdline_store.from_cmdline(
                ['vm.name={}'.format(self.vm_name)])
            try:
                kls_name = conf.get('vm.class', convert=False)
            except config_errors.OptionMandatoryValueError:
                raise errors.UciVmsError('"{name}" must define vm.class.',
                                         name=self.vm_name)
            # Even if defined, the value may be wrong
            try:
                kls = conf.get('vm.class')
            except config_errors.OptionMandatoryValueError:
                raise errors.InvalidVmClass(self.vm_name, kls_name)
            self.vm = kls(conf)
        return self.options


class Config(VmCommand):

    name = 'config'
    description = 'Manage a virtual machine configuration.'

# FIXME: The full help should be the following but that doesn't fit well with
# the Help command for now -- vila 2014-03-26

# Display the active value for option NAME.
#
# If NAME is not given, --all .* is implied (all options are displayed).
#
# If --all is specified, NAME is interpreted as a regular expression and all
# matching options are displayed (without resolving option references in the
# value). The active value is the first one displayed for each option.
#
# NAME=value without spaces sets the value.
#
# --remove NAME remove the option definition.

    def __init__(self, **kwargs):
        super(Config, self).__init__(**kwargs)
        self.parser.add_argument(
            'name', nargs='?', help='''The option name.''')
        self.parser.add_argument(
            '--remove', '-r', action="store_true", help='Remove the option.'),
        self.parser.add_argument(
            '--all', '-a', action="store_true",
            help='Display all the defined values for the matching options.'),

    def parse_args(self, args):
        # Work around argparse not providing a proper way to mix positional
        # arguments and options. If an option arg is used, the 'name'
        # positional arg is not set and ends up in 'rest'.
        self.options, rest = self.parser.parse_known_args(
            args, self.out, self.err)
        self.parse_vm_arg()
        o = self.options
        if o.all or o.remove:
            # Get 'name' back from 'rest' if available
            if rest:
                o.name = rest.pop(0)
        # If we still have args left, it's an error
        if rest:
            err_orig = sys.stderr
            try:
                sys.stderr = self.err
                msg = 'unrecognized arguments: {}'
                self.parser.error(msg.format(' '.join(rest)))
            finally:
                sys.stderr = err_orig
        if o.remove:
            if o.all:
                raise errors.UciVmsError(
                    '--remove and --all are mutually exclusive.')
            if o.name is None:
                raise errors.UciVmsError(
                    '--remove expects an option to remove.')
        if o.name is not None and '=' in o.name:
            if o.all:
                raise errors.UciVmsError('Only one option can be set.')
        return o

    def list(self, name):
        # Display the value as a string
        value = self.vm.conf.get(name, expand=True, convert=False)
        if value is not None:
            # No quoting needed (for now)
            self.out.write(value)
        else:
            raise errors.ConfigOptionNotFound(name)

    def list_matching(self, re_str):
        not_a_section_id = object()
        cur_store_id = None
        cur_section_id = not_a_section_id
        names = re.compile(re_str)
        for (store, section, name, value) in self.vm.conf.iter_options():
            if names.search(name) is None:
                # not matching option, ignore
                continue
            if cur_store_id != store.id:
                # Explain where the options are defined
                self.out.write('{}:\n'.format(store.id,))
                cur_store_id = store.id
                cur_section_id = not_a_section_id
            if (cur_section_id != section.id):
                if section.id is None:
                    section_displayed = ''
                else:
                    section_displayed = section.id
                # Display the section id as it appears in the store
                self.out.write('  [{}]\n'.format(section_displayed))
            cur_section_id = section.id
            # No quoting needed (for now)
            self.out.write('  {} = {}\n'.format(name, value))

    def run(self):
        o = self.options
        if o.remove:
            try:
                self.vm.conf.remove(o.name)
            except config_errors.NoSuchConfigOption as e:
                raise errors.ConfigOptionNotFound(e.name)
        elif o.name is not None and '=' in o.name:
            # Set the option value
            name, value = o.name.split('=', 1)
            self.vm.conf.set(name, value)
        else:
            # List the options
            if o.name is None:
                # Defaults to all options
                o.name = '.*'
                o.all = True
            if o.all:
                self.list_matching(o.name)
            else:
                self.list(o.name)


command_registry.register(Config)


class Setup(VmCommand):

    name = 'setup'
    description = 'Setup a virtual machine.'

    def __init__(self, **kwargs):
        super(Setup, self).__init__(**kwargs)
        self.parser.add_argument(
            '--download', '-d', action="store_true",
            help='Force download of the required image.')
        self.parser.add_argument(
            '--ssh-keygen', '-k', action="store_true",
            help='Generate the ssh keys for the machine).')

    def run(self):
        if self.options.download:
            self.vm.download(force=True)
        if self.options.ssh_keygen:
            self.vm.ssh_keygen(force=True)

        state = self.vm.state()
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in('shut off', 'STOPPED'):
            self.vm.teardown()
        elif state in ('running', 'RUNNING'):
            raise errors.VmRunning(self.vm_name)
        self.vm.setup()
        return 0


command_registry.register(Setup)


class Status(VmCommand):

    name = 'status'
    description = 'Display the status of a virtual machine.'

    def run(self):

        state = self.vm.state()
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('shut off', 'STOPPED'):
            state = 'STOPPED'
        elif state in ('running', 'RUNNING'):
            state = 'RUNNING'
        self.out.write(state + '\n')
        return 0


command_registry.register(Status)


class Start(VmCommand):

    name = 'start'
    description = 'Start an existing virtual machine.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('running', 'RUNNING'):
            raise errors.VmRunning(self.vm_name)
        self.vm.start()
        return 0


command_registry.register(Start)


class Shell(VmCommand):

    name = 'shell'
    description = 'Run a command on an existing virtual machine.'

    def __init__(self, **kwargs):
        super(Shell, self).__init__(**kwargs)
        self.parser.add_argument(
            'command', help='The command to run on the vm.', nargs='?')
        self.parser.add_argument(
            'args', help='The arguments for the command to run on the vm.',
            nargs=argparse.REMAINDER)

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        retcode = self.vm.shell(self.options.command, *self.options.args)
        return retcode


command_registry.register(Shell)


class Stop(VmCommand):

    name = 'stop'
    description = 'Stop an existing virtual machine.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state not in ('running', 'RUNNING'):
            raise errors.VmNotRunning(self.vm_name)
        self.vm.stop()
        return 0


command_registry.register(Stop)


class Teardown(VmCommand):

    name = 'teardown'
    description = 'Teardown a virtual machine.'

    def run(self):
        state = self.vm.state()
        if state is None:
            raise errors.VmUnknown(self.vm_name)
        # FIXME: states need to be defined uniquely across the various vms
        # implementations -- vila 2014-01-17
        if state in ('running', 'RUNNING'):
            raise errors.VmRunning(self.vm_name)
        self.vm.teardown()
        return 0


command_registry.register(Teardown)


def run(args=None):
    if args is None:
        args = sys.argv[1:]
    if not args:
        cmd = Help()
    else:
        cmd_name = args[0]
        args = args[1:]
        try:
            cmd_class = command_registry.get(cmd_name)
            cmd = cmd_class()
        except KeyError:
            cmd = Help()
            args = [cmd_name]
    try:
        cmd.parse_args(args)
        sys.exit(cmd.run())
    except (errors.BaseError, config_errors.ConfigError) as e:
        sys.stderr.write('ERROR: {!r}\n'.format(e))
        sys.exit(-1)
